#include "pch.h"
// ***************************************************************************
// This comes from the pc-ble-driver: examples\heart_rate_collector\main.c
// ***************************************************************************


/*
 * copyright (c) 2012 - 2018, nordic semiconductor asa
 * all rights reserved.
 *
 * redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. redistributions in binary form, except as embedded into a nordic
 *    semiconductor asa integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /**@example examples/heart_rate_collector
  *
  * @brief Heart Rate Collector Sample Application main file.
  *
  * This file contains the source code for a sample application that acts as a BLE Central device.
  * This application scans for a Heart Rate Sensor device and reads it's heart rate data.
  * https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.service.heart_rate.xml
  *
  * Structure of this file
  * - Includes
  * - Definitions
  * - Global variables
  * - Global functions
  * - Event functions
  * - Event dispatcher
  * - Main
  */

  /** Includes */
#include "message_queue.h"
#include "windows.h"
#include "process.h" // declares _beginthreadex
#include <vector>
#include <map>
#include <set>
#include <mutex>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include "NordicErrors.h"

static std::string GetCurrentDateTimeString()
{
	using namespace std::chrono;

	// get current time
	auto now = system_clock::now();

	// get number of milliseconds for the current second
	// (remainder after division into seconds)
	auto ms = duration_cast<milliseconds>(now.time_since_epoch()) % 1000;

	// convert to std::time_t in order to convert to std::tm (broken time)
	auto timer = system_clock::to_time_t(now);

	// convert to broken time
	std::tm bt = *std::localtime(&timer);

	std::ostringstream oss;

	oss << std::put_time(&bt, "%Y-%m-%d %H:%M:%S"); // HH:MM:SS
	oss << '.' << std::setfill('0') << std::setw(3) << ms.count();

	return oss.str();
}

bgsdk::details::MessageQueue message_queue_;

/** Definitions */
#define DEFAULT_BAUD_RATE 1000000 /**< The baud rate to be used for serial communication with nRF5 device. */
#define MAX_CONNECTION_COUNT 5

#ifdef _WIN32
#define DEFAULT_UART_PORT_NAME "COM1"
#endif
#ifdef __APPLE__
#define DEFAULT_UART_PORT_NAME "/dev/tty.usbmodem00000"
#endif
#ifdef __linux__
#define DEFAULT_UART_PORT_NAME "/dev/ttyACM0"
#endif

enum
{
	UNIT_0_625_MS = 625,  /**< Number of microseconds in 0.625 milliseconds. */
	UNIT_1_25_MS = 1250, /**< Number of microseconds in 1.25 milliseconds. */
	UNIT_10_MS = 10000 /**< Number of microseconds in 10 milliseconds. */
};

#define MSEC_TO_UNITS(TIME, RESOLUTION) (((TIME) * 1000) / (RESOLUTION))

#define SCAN_INTERVAL 0x00A0 /**< Determines scan interval in units of 0.625 milliseconds. */
#define SCAN_WINDOW   0x0050 /**< Determines scan window in units of 0.625 milliseconds. */
#define SCAN_TIMEOUT  0x0    /**< Scan timeout between 0x01 and 0xFFFF in seconds, 0x0 disables timeout. */

#define MIN_CONNECTION_INTERVAL         MSEC_TO_UNITS(7.5, UNIT_1_25_MS) /**< Determines minimum connection interval in milliseconds. */
#define MAX_CONNECTION_INTERVAL         MSEC_TO_UNITS(7.5, UNIT_1_25_MS) /**< Determines maximum connection interval in milliseconds. */
#define SLAVE_LATENCY                   0                                /**< Slave Latency in number of connection events. */
#define CONNECTION_SUPERVISION_TIMEOUT  MSEC_TO_UNITS(4000, UNIT_10_MS)  /**< Determines supervision time-out in units of 10 milliseconds. */

#define BLE_UUID_HEART_RATE_MEASUREMENT_CHAR 0x2A37 /**< Heart Rate Measurement characteristic UUID. */
#define BLE_UUID_CCCD                        0x2902
#define BLE_CCCD_NOTIFY                      0x01

// We do a keepalive of phone's with Imprivata ID installed by polling the 'feature' characteristic
// From OneSign source: const Uuid FEATURE_AUTHENTICATION_CHARACTERISTIC_UUID(L"57862960-5B0F-4CE0-A255-30997B420C79");
const char* IMPRIVATA_ID_FEATURE_CHARACTERISTIC_UUID_NO_DASHES = "578629605B0F4CE0A25530997B420C79";// 32 hex characters == 16 bytes == 128 bits

// From OneSign source, the challenge/response characteristic UUIDs
const char* IMPRIVATA_ID_RESPONSE_CHARACTERISTIC_UUID_NO_DASHES = "0B9DB0E5C5B34043829444D8112ADC54";
const char* IMPRIVATA_ID_CHALLENGE_CHARACTERISTIC_UUID_NO_DASHES = "7198C97A914A432DB8280EEA0E2B65FC";

#define STRING_BUFFER_SIZE 50

typedef struct
{
	uint8_t *     p_data;   /**< Pointer to data. */
	uint16_t      data_len; /**< Length of data. */
} data_t;

bool m_DoScanning = true;
/** Global variables */
std::mutex m_MutexEnforceSingleCallToNordic;
static unsigned long long	m_timestampOpenNordicDevice = 0;// tick count when we opened the Nordic device
static unsigned long long	m_timestampLastAdvertisement = 0;
static unsigned long long	m_timestampLastScanStart = 0;
static uint16_t    m_hrm_char_handle = 0;
static uint16_t    m_hrm_cccd_handle = 0;

int					m_totalAdvertisementsReceived = 0;
int					m_totalAdvertisementsReceivedWith128BitUUIDs = 0;

std::set<std::string> m_all128bitUUIDDevices;


struct ConnectedCharacteristic
{
	uint16_t	m_DescriptionHandle;
	uint16_t	m_ValueHandle;
	bool		m_bGotDescriptionDetails;
	std::string	m_ValueHandleUUID;

	ConnectedCharacteristic(uint16_t desc, uint16_t val) : m_DescriptionHandle(desc), m_ValueHandle(val), m_bGotDescriptionDetails(false){}
};
struct ConnectedService
{
	ConnectedService(uint8_t t, uint16_t shortUUID, const ble_gattc_handle_range_t& handle_range)
	{
		m_uuid_type = t;
		m_uuid_value.m_uuid_16bit = shortUUID;
		m_handle_range = handle_range;
	}
	ConnectedService(uint8_t t, const ble_gattc_handle_range_t& handle_range)
	{
		m_uuid_type = t;
		// m_uuid_value is not set
		m_handle_range = handle_range;
	}
	ConnectedService(uint8_t t, const ble_uuid128_t& longUUID, const ble_gattc_handle_range_t& handle_range)
	{
		m_uuid_type = t;
		m_uuid_value.m_uuid_128_bit = longUUID;
		m_handle_range = handle_range;
	}
	uint8_t     m_uuid_type; /**< UUID type, see @ref BLE_UUID_TYPES. If type is @ref BLE_UUID_TYPE_UNKNOWN, the value of uuid is undefined. */
	union // if m_uuid_type==BLE_UUID_TYPE_UNKNOWN, then this union has no value
	{
		uint16_t		m_uuid_16bit; //set if m_uuid_type==BLE_UUID_TYPE_BLE
		ble_uuid128_t	m_uuid_128_bit;// set if m_uuid_type = BLE_UUID_TYPE_VENDOR_BEGIN

	} m_uuid_value;
	std::string m_uuid_string;

	ble_gattc_handle_range_t	m_handle_range;

	std::vector<ConnectedCharacteristic>	m_Characteristics;
};

struct ConnectedDevice
{
	ConnectedDevice(const std::string& uuid) : m_UUID(uuid), m_ConnectionHandle(BLE_CONN_HANDLE_INVALID) {}

	std::string						m_UUID;// we only connect to devices that advertise a UUID we are looking for
	uint16_t						m_ConnectionHandle;
	std::vector<ConnectedService>	m_Services;
	int8_t							m_lastRSSI{-127};
	int								m_numberRSSIEvents{0};
};
static std::map<std::string/*UUID*/, ConnectedDevice*>	ConnectedDevices;

// When we decide to connect to a device, we have these steps: connect, get services, get characteristic list, get characteristic description meta-data
// 'm_DeviceBeingDiscovered' will be non-null during that entire process
static ConnectedDevice*	m_DeviceBeingDiscovered = nullptr;

// Event that is set when we receive event BLE_GATTC_EVT_READ_RSP
static HANDLE hEvent_BLE_GATTC_EVT_READ_RSP = 0;

// Event that is set whenever we receive event BLE_GATTC_EVT_WRITE_RSP
static HANDLE hEvent_BLE_GATTC_EVT_WRITE_RSP = 0;

static HANDLE hEvent_current_GATTC = 0;// set to hEvent_BLE_GATTC_EVT_READ_RSP when we're doing a read, hEvent_BLE_GATTC_EVT_WRITE_RSP when we're doing a write, otherwise 0

static bool current_GATTC_timed_out = false;// set to true if we get event BLE_GATTC_EVT_TIMEOUT

// Event that is set whenever we receive event BLE_GAP_EVT_DISCONNECTED
static HANDLE hEvent_BLE_GAP_EVT_DISCONNECTED = 0;

// Event that is set when a connect request is done, either connected or timed out
static HANDLE hEvent_BLE_GAP_EVT_CONNECT_Completed = 0;
static bool current_CONNECT_timed_out;

void remove_connected_device(uint16_t conn_handle)
{
	for(auto dev : ConnectedDevices)
	{
		if(dev.second->m_ConnectionHandle == conn_handle)
		{
			dev.second->m_ConnectionHandle = BLE_CONN_HANDLE_INVALID;
			dev.second->m_Services.clear();
			// leave the UUID, it's what identifies the entry in ConnectedDevices
		}
	}
}
void update_connected_device_rssi_changed(uint16_t conn_handle, int8_t rssi)
{
	for (auto dev : ConnectedDevices)
	{
		if (dev.second->m_ConnectionHandle == conn_handle)
		{
			dev.second->m_numberRSSIEvents++;
			dev.second->m_lastRSSI = rssi;
		}
	}
}

void log_connected_device_status()
{
	for (auto dev : ConnectedDevices)
	{
		if (dev.second->m_ConnectionHandle == BLE_CONN_HANDLE_INVALID)
		{
			printf("%s: %s not connected\n", GetCurrentDateTimeString().c_str(), dev.second->m_UUID.c_str());
		}
		else
		{
			printf("%s: %s connection %d #RSSI events = %d last RSSI = %d\n", GetCurrentDateTimeString().c_str(), dev.second->m_UUID.c_str(), dev.second->m_ConnectionHandle, dev.second->m_numberRSSIEvents, dev.second->m_lastRSSI);
		}
		fflush(stdout);
	}
}
void log_all_128bitUUID_devices_seen()
{
	printf("%s: all devices: ", GetCurrentDateTimeString().c_str());
	for(auto dev : m_all128bitUUIDDevices)
	{
		printf("%s ", dev.c_str());
	}
	printf("\n");
	fflush(stdout);
}



static adapter_t * m_adapter = NULL;
static bool m_bShuttingDown = false;

#if NRF_SD_BLE_API >= 5
static uint32_t    m_config_id = 1;
#endif

#if NRF_SD_BLE_API >= 6
static uint8_t     mp_data[100] = { 0 };
static ble_data_t  m_adv_report_buffer;
#endif


static const ble_gap_conn_params_t m_connection_param =
{
	(uint16_t)MIN_CONNECTION_INTERVAL,
	(uint16_t)MAX_CONNECTION_INTERVAL,
	(uint16_t)SLAVE_LATENCY,
	(uint16_t)CONNECTION_SUPERVISION_TIMEOUT
};

// Added for Imprivata test program to find phone with given UUID
#include "Calls_to_pc_ble_driver.h"
static int m_iLoggingLevel = 0;// higher number get more verbose logging
const int UUID_BYTE_LENGTH = 16;// 16 bytes is 128 bits
const int UUID_CHARACTER_LENGTH_NONULL = UUID_BYTE_LENGTH * 2;// each byte takes two hex digits when formatted as a string
static bool UUID_to_string_convert(uint16_t rawDataLen, const uint8_t* rawData, char* string_buffer, size_t string_buffer_nchars)
{
	if (rawDataLen != UUID_BYTE_LENGTH)
	{
		//OutputDebugString("ignoring UUID that is not 128");//= ");OutputDebugString(rawData.data_len);advert_data_buffer OutputDebugString(" that is not canonical length = ", UUID_BYTE_LENGTH);
		return false;
	}
	// The binary order is the opposite of the canonical order we use
	for (int iByte = 0; iByte < rawDataLen; ++iByte)
		sprintf_s(string_buffer + 2 * iByte, string_buffer_nchars - 2 * iByte, "%02x", rawData[rawDataLen-iByte-1]);

	return true;
}

static void force_disconnect(uint16_t connectionHandle, bool bWaitForDisconnectToFinish)
{
	// Should receive hEvent_BLE_GAP_EVT_DISCONNECTED
	::ResetEvent(hEvent_BLE_GAP_EVT_DISCONNECTED);
	uint32_t disconnectCode;
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		disconnectCode = sd_ble_gap_disconnect(m_adapter, connectionHandle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
	}
	if (disconnectCode == NRF_SUCCESS)
	{
		if (m_iLoggingLevel >= 2)
			printf("%s: Thread %d: sd_ble_gap_disconnect succeeded handle %d\n",
			GetCurrentDateTimeString().c_str(),
			::GetCurrentThreadId(),
			connectionHandle);

		if (bWaitForDisconnectToFinish)
		{
			const DWORD millisecondsToWaitForDisconnect = 65000;
			DWORD dWait = ::WaitForSingleObject(hEvent_BLE_GAP_EVT_DISCONNECTED, millisecondsToWaitForDisconnect);
			if (dWait == WAIT_OBJECT_0)
			{
				if (m_iLoggingLevel >= 1)
					printf("%s: Thread %d: got expected BLE_GAP_EVT_DISCONNECTED after calling sd_ble_gap_disconnect\n", GetCurrentDateTimeString().c_str(), ::GetCurrentThreadId());
			}
			else if (dWait == WAIT_TIMEOUT)
			{
				printf("%s: Thread %d: ERROR: never got BLE_GAP_EVT_DISCONNECTED after waiting %d milliseconds.\n",
					GetCurrentDateTimeString().c_str(),
					::GetCurrentThreadId(),
					millisecondsToWaitForDisconnect
				);
				printf("Exit test program immediately\n");
				::ExitProcess(0/*exit code*/);
			}
			else
			{
				printf("ERROR: WaitForSingleObject returned unexpected value %d while waiting for BLE_GAP_EVT_DISCONNECTED\n", dWait);
			}
		}
	}
	else
	{
		printf("%s: ERROR: sd_ble_gap_disconnect failed handle %d error %s\n", GetCurrentDateTimeString().c_str(), connectionHandle, bgsdk::Nordic::NordicErrorCodeToMessage(disconnectCode).c_str());
	}
}


static void force_disconnect_on_all_connections()
{
	for (auto& con : ConnectedDevices)
	{
		if (con.second->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
		{
			force_disconnect(con.second->m_ConnectionHandle, true/*pass 'true' to wait for events indicating disconnect is done*/);
		}
	}
}

static bool enable_notifications_for_response_characteristic(std::string& authenticationServiceUUID, ConnectedDevice* conn, uint16_t charHandleResponse, bool bEnable)
{
	if (conn->m_ConnectionHandle == BLE_CONN_HANDLE_INVALID)
		return false;

	bool bSuccess = false;
	{
		// Enable the response
		ble_gattc_write_params_t write_params;
		uint8_t                  challenge_value[19] = { 0 };// from OneSign source, challenge is 19 bytes, the value does not matter for this test
		uint8_t bytes[2] = { 0x00, 0x00 };
		if (bEnable)
			bytes[0] = 0x01;

		write_params.handle = charHandleResponse + 1/*the handle for notification is +1 from the characteristic*/;
		write_params.len = 2;
		write_params.p_value = bytes;
		write_params.write_op = BLE_GATT_OP_WRITE_REQ; // This gets BLE_GATTC_EVT_WRITE_RSP
		write_params.offset = 0;

		if (hEvent_current_GATTC)
		{
			printf("%s: WARNING skip sd_ble_gattc_write when there is already a GATTC operation in progress - this should not happen\n",
				GetCurrentDateTimeString().c_str());
			return false;
		}
		current_GATTC_timed_out = false;
		hEvent_current_GATTC = hEvent_BLE_GATTC_EVT_WRITE_RSP;

		int32_t err_code;
		{
			std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
			err_code = sd_ble_gattc_write(m_adapter, conn->m_ConnectionHandle, &write_params);
		}
		if (err_code != NRF_SUCCESS)
		{
			hEvent_current_GATTC = 0;// since we won't wait on event
			printf("%s: ERROR requesting %s for phone %s conn_handle = %d, error %s\n",
				GetCurrentDateTimeString().c_str(),
				(bEnable ? "enable" : "disable"),
				authenticationServiceUUID.c_str(), conn->m_ConnectionHandle, bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		}
		else
		{
			if (m_iLoggingLevel >= 4)
			{
				printf("%s: Thread %d: %s notify for %s conn %d response handle 0x%04X\n",
					GetCurrentDateTimeString().c_str(),
					::GetCurrentThreadId(),
					(bEnable ? "enable" : "disable"),
					authenticationServiceUUID.c_str(),
					conn->m_ConnectionHandle,
					charHandleResponse);
			}
			// We should get event BLE_GATTC_EVT_WRITE_RSP.
			// The maximum time according to the BLE spec is 30 seconds, after which pc-ble-driver gives us a timeout event.
			const DWORD millisecondsToWaitForWriteEvent = 65000;
			DWORD dWait = ::WaitForSingleObject(hEvent_BLE_GATTC_EVT_WRITE_RSP, millisecondsToWaitForWriteEvent);
			hEvent_current_GATTC = 0;// done waiting on event
			if (dWait == WAIT_OBJECT_0)
			{
				if (current_GATTC_timed_out)
				{
					printf("%s: sd_ble_gattc_write for characteristic notification got BLE_GATTC_EVT_TIMEOUT or BLE_GAP_EVT_DISCONNECTED (discard this connection and re-connect)\n", GetCurrentDateTimeString().c_str());
					if(conn->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)// check, in case we got a disconnect event
						force_disconnect(conn->m_ConnectionHandle, true/*do wait for disconnect event*/);
				}
				else 
				{
					bSuccess = true;
					if (m_iLoggingLevel >= 4)
					{
						printf("%s: Thread %d: got expected BLE_GATTC_EVT_WRITE_RSP\n",
							GetCurrentDateTimeString().c_str(),
							::GetCurrentThreadId()
						);
					}
				}
			}
			else if (dWait == WAIT_TIMEOUT)
			{
				printf("%s: Thread %d: ERROR: never got BLE_GATTC_EVT_WRITE_RSP or BLE_GAP_EVT_DISCONNECTED after waiting %d milliseconds.\n",
					GetCurrentDateTimeString().c_str(),
					::GetCurrentThreadId(),
					millisecondsToWaitForWriteEvent
				);
				if (conn->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
				{
					printf("Before exiting, disconnect and see if we get a disconnect event\n");
					force_disconnect(conn->m_ConnectionHandle, true/*do wait for disconnect event, we want to know if it exists*/);
				}
				printf("Exit test program immediately\n");
				::ExitProcess(0/*exit code*/);
			}
			else
			{
				printf("ERROR: WaitForSingleObject returned unexpected value %d while waiting for BLE_GATTC_EVT_WRITE_RSP\n", dWait);
			}
		}
	}
	return bSuccess;
}



void do_challenge_response_to_connected_phone(std::string& authenticationServiceUUID, ConnectedDevice* conn, uint16_t charHandleChallenge, uint16_t charHandleResponse)
{
	if (!enable_notifications_for_response_characteristic(authenticationServiceUUID, conn, charHandleResponse, true))
		return;// don't keep doing challenge/response if this fails

	// Setting the challenge
	else if(conn->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
	{
		ble_gattc_write_params_t write_params;
		uint8_t                  challenge_value[19] = { 0 };// from OneSign source, challenge is 19 bytes, the value does not matter for this test

		write_params.handle = charHandleChallenge;
		write_params.len = 19;
		write_params.p_value = challenge_value;
		write_params.write_op = BLE_GATT_OP_WRITE_CMD; // The example we started with had BLE_GATT_OP_WRITE_REQ;
		write_params.offset = 0;

		int32_t err_code;
		{
			std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
			err_code = sd_ble_gattc_write(m_adapter, conn->m_ConnectionHandle, &write_params);
		}

		if (err_code != NRF_SUCCESS)
		{
			printf("%s: ERROR setting challenge for phone %s conn_handle = %d, error %s\n", 
				GetCurrentDateTimeString().c_str(), 
				authenticationServiceUUID.c_str(), 
				conn->m_ConnectionHandle,
				bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
			return;// don't keep doing challenge/response if this fails
		}
		else
		{
			if (m_iLoggingLevel >= 4)
				printf("%s: Thread %d: success setting challenge for phone %s conn_handle = %d challenge char 0x%04X\n",
				GetCurrentDateTimeString().c_str(),
				::GetCurrentThreadId(),
				authenticationServiceUUID.c_str(),
				conn->m_ConnectionHandle,
				charHandleChallenge);
			// We should get event BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE. This test program has no reason to check if we get it

			// The response will be 14 separate notifications. For now I'm not going to check for the response being received,
			// but I"ll wait so we receive some.
			::Sleep(3000);
			enable_notifications_for_response_characteristic(authenticationServiceUUID, conn, charHandleResponse, false);
		}
	}
}


void do_challenge_response_to_connected_phones()
{
	int iConnectedDevices = 0;

	{
		{
			// can we keep other connections alive while one is being discovered? I sure hope so, otherwise they may time out
			if (m_DeviceBeingDiscovered)
				return;

			// Try to contact every open connection
			struct ConnectionAndCharacteristic {
				std::string m_authenticationServiceUUID;
				ConnectedDevice* m_conn;
				uint16_t m_charHandleChallenge;
				uint16_t m_charHandleResponse;

				ConnectionAndCharacteristic(std::string& authService, ConnectedDevice* con, uint16_t chChallenge, uint16_t chResponse) :
					m_authenticationServiceUUID(authService), m_conn(con), m_charHandleChallenge(chChallenge), m_charHandleResponse(chResponse) {}
			};
			std::vector< ConnectionAndCharacteristic> openConnections;

			for (auto dev : ConnectedDevices)
			{
				if (dev.second->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
				{
					iConnectedDevices++;
					uint16_t challengeHandle = 0;// the Imprivata ID 'challenge' characteristic, whose UUID is known to us
					uint16_t responseHandle = 0;// the Imprivata ID 'response' characteristic, whose UUID is known to us
					for (auto srv : dev.second->m_Services)
					{
						if (srv.m_uuid_string == dev.second->m_UUID)// found the IID authentication service
						{
							for (auto ch : srv.m_Characteristics)
							{
								if (_stricmp(ch.m_ValueHandleUUID.c_str(), IMPRIVATA_ID_CHALLENGE_CHARACTERISTIC_UUID_NO_DASHES) == 0)
								{
									challengeHandle = ch.m_ValueHandle;
									if (m_iLoggingLevel >= 4)
										printf("authentication service %s 'challenge' characteristic value 0x%04X handle\n",
											srv.m_uuid_string.c_str(),
											ch.m_ValueHandle);
								}
								else if (_stricmp(ch.m_ValueHandleUUID.c_str(), IMPRIVATA_ID_RESPONSE_CHARACTERISTIC_UUID_NO_DASHES) == 0)
								{
									responseHandle = ch.m_ValueHandle;
									if (m_iLoggingLevel >= 4)
										printf("authentication service %s 'response' characteristic value 0x%04X handle\n",
											srv.m_uuid_string.c_str(),
											ch.m_ValueHandle);
								}
							}
						}
					}

					if (challengeHandle == 0 || responseHandle == 0)
					{
						printf("phone %s connection_handle = %d could not find 'challenge' and 'respnose' characteristics\n",
							dev.second->m_UUID.c_str(),
							dev.second->m_ConnectionHandle);
						fflush(stdout);
					}
					else
					{
						openConnections.push_back(ConnectionAndCharacteristic(dev.second->m_UUID, dev.second, challengeHandle, responseHandle));
					}
				}
			}
			for (auto con : openConnections)
			{
				do_challenge_response_to_connected_phone(con.m_authenticationServiceUUID, con.m_conn, con.m_charHandleChallenge, con.m_charHandleResponse);
			}
		}
	}

	if (iConnectedDevices == 0)
	{
		if (m_iLoggingLevel >= 4)
			printf("no devices are connected, cannot test challenge/response\n");
	}
}


/** Global functions */
/**@brief Function for setting the advertisement data.
 *
 * @details Sets the full device name and its available BLE services in the advertisement data.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint8_t                  m_adv_handle = BLE_GAP_ADV_SET_HANDLE_NOT_SET;// original sample started at 0, but doc says to use BLE_GAP_ADV_SET_HANDLE_NOT_SET "to configure a new advertising set"

#define ADVERTISING_INTERVAL_40_MS 64  /**< 0.625 ms = 40 ms */
#define ADVERTISING_INTERVAL_100_MS 160  /* 160* 0.625 ms = 100 ms, which is the minimum for non-connectable advertising */
#define ADVERTISING_TIMEOUT_3_MIN  180 /**< 1 sec = 3 min */


static const unsigned	ADVERT_DATA_SIZE = 30;
const unsigned			ADVERT_UUID_OFFSET = 9;	  	// UUID is after 9 preliminary bytes
const unsigned			ADVERT_UUID_SIZE = 16;
const unsigned			ADVERT_MAJOR_LSB_OFFSET = 26;	  	// Major field is after the UUID, with most-significant-byte followed by least-significant-byte
const unsigned			ADVERT_MINOR_LSB_OFFSET = 28;

static void check_advert_buffer_for_corruption(const uint8_t* advert_data_buffer, const char* pContext)
{
	if (advert_data_buffer[0] != 2 || advert_data_buffer[1] != 1 || advert_data_buffer[2] != 6 || advert_data_buffer[3] != 26)
	{
		printf("%s: advertising data buffer has been corrupted. Initial bytes should be 02 01 06 1A but are %02x %02x %02x %02x\n",
			pContext,
			advert_data_buffer[0],
			advert_data_buffer[1],
			advert_data_buffer[2],
			advert_data_buffer[3]);
		return;
	}
}
static uint32_t advertising_data_setup()
{
	// I think the advertising sample code came from pc-ble-driver examples\heart_rate_monitor\main.c

	// These data structures passed to pc-ble-driver need to be non-static data structures, 
	// because they get partially overwritten by pc-ble-driver (at least advert_data_buffer does).

	uint8_t			advert_data_buffer[ADVERT_DATA_SIZE] =// copied from OneSign source dev\bluetooth_sdk\bgsdk\src\AdvertisingData.cpp
	{
					 0x02,		// length of data (Flags)
					 0x01, 		// AD type = Flags
					 0x06, 		// bit mask - See Bluetooth Specification Version 4.0 Vol 3 - Part C - Chapter 18
								//  bit 0 - LE Limited Discoverable Mode
								//  bit 1 - LE General Discoverable Mode
								//  bit 2 - BR/EDR Not Supported
								//  bit 3 - Simultaneous LE and BR/EDR to same device (controller)
								//  bit 4 - Simultaneous LE and BR/EDR to same device (host)
								//  bit 5-7 reserved
					 0x1a,		// length of AD data
					 0xff,		// manufacturer specific data
					 0x4c, 0x00, // manufacturer code
					 0x02, 		// iBeacon device type
					 0x15,		// iBeacon data length
					 // UUID that's the same for all Imprivata ID phones (generic UUID for waking the IID application)
					 0XCE, 0X99, 0X3B, 0XE7, 0XC9, 0X40, 0X45, 0XEB, 0X87, 0X48, 0XA4, 0XE6, 0X06, 0X21, 0XCF, 0XBA,
					 // Major
					 0x00, 0x01,
					 // Minor
					 0x00, 0x01,
					 // TXPower @ 1 meter (Bluegiga dongle calculated using iPhone+Locate from appstore
					 0xC0  // 0xC0 == -64
	};
	check_advert_buffer_for_corruption(advert_data_buffer, "before sd_ble_gap_adv_set_configure");

	uint32_t err_code;

	// Rotate the UUID value that we beacon
	{
		static int iMajor = 1;
		static int iMinor = 1;
		static uint8_t iBeaconUUIDByte = 0xDE;
		iBeaconUUIDByte = (iBeaconUUIDByte+1)&0xFF;// iterate across all byte values
		for(int iOffset = ADVERT_UUID_OFFSET; iOffset < ADVERT_UUID_OFFSET+ ADVERT_UUID_SIZE; ++iOffset)
			advert_data_buffer[iOffset] = iBeaconUUIDByte;

		// Rotate major/minor from range 1 through 20
		if(iMajor <= iMinor)
			++iMajor;
		else
			++iMinor;

		if(iMajor > 20 || iMinor > 20)
		{
			iMajor = 1;
			iMinor = 1;
		}
		advert_data_buffer[ADVERT_MAJOR_LSB_OFFSET] = iMajor;
		advert_data_buffer[ADVERT_MINOR_LSB_OFFSET] = iMinor;
	}

	// No scan response.

#if NRF_SD_BLE_API >= 6
	static ble_gap_adv_properties_t adv_properties;
	adv_properties.type = BLE_GAP_ADV_TYPE_NONCONNECTABLE_NONSCANNABLE_UNDIRECTED;//BLE_GAP_ADV_TYPE_NONCONNECTABLE_SCANNABLE_UNDIRECTED;// Example I copied used connectable beacon BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
	adv_properties.anonymous = 0;// This is used only for 'extended' types BLE_GAP_ADV_TYPE_EXTENDED_*
	adv_properties.include_tx_power = 0;// This is used only for 'extended' types BLE_GAP_ADV_TYPE_EXTENDED_*

	// Because we set both 'duration' and 'max_adv_evts' to zero, advertising won't auto-stop, 
	// so we'll never get event BLE_GAP_EVT_ADV_SET_TERMINATED according to https://devzone.nordicsemi.com/f/nordic-q-a/33106/bug-ble_gap_evt_adv_set_terminated-is-never-raised
	static ble_gap_adv_params_t     m_adv_params;
	m_adv_params.properties = adv_properties;
	m_adv_params.filter_policy = BLE_GAP_ADV_FP_ANY;
	m_adv_params.duration = BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED;
	m_adv_params.p_peer_addr = NULL;
	m_adv_params.interval = ADVERTISING_INTERVAL_100_MS;// I've read non-connectable advertising has 100ms minimum interval
	m_adv_params.max_adv_evts = 0;// 0 means no limit. When I tested setting this to 1 I got event BLE_GAP_EVT_ADV_SET_TERMINATED as expected, but then got "Failed to decode event, error code is 14/0xe."
	m_adv_params.primary_phy = BLE_GAP_PHY_AUTO;
	m_adv_params.secondary_phy = BLE_GAP_PHY_AUTO;

	// m_adv_params.channel_mask is a 5-byte (thus 40 bit) bitmask of channels.
	// "If a bit is set to 1, the channel is not used." Therefore setting this to all zeros means to use all channels.
	m_adv_params.channel_mask[0] = 0;
	m_adv_params.channel_mask[1] = 0;
	m_adv_params.channel_mask[2] = 0;
	m_adv_params.channel_mask[3] = 0;
	m_adv_params.channel_mask[4] = 0;
	m_adv_params.set_id = 0;// Don't have to set this. Header file says "This parameter will be ignored when ble_gap_adv_properties_t::type is not an extended advertising type."
	m_adv_params.scan_req_notification = 0;// Don't have to set this. header file says "This parameter will be ignored when ble_gap_adv_properties_t::type is a non - scannable	advertising type."

	static ble_data_t		adv_data;
	adv_data.p_data = advert_data_buffer;
	adv_data.len = sizeof(advert_data_buffer);

	static ble_data_t scan_rsp_data;
	scan_rsp_data.p_data = NULL;
	scan_rsp_data.len = 0;

	static ble_gap_adv_data_t		m_adv_data;
	m_adv_data.adv_data = adv_data;
	m_adv_data.scan_rsp_data = scan_rsp_data;

	// This is the place where m_adv_handle gets set. It is later used to start/stop advertising.
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_ble_gap_adv_set_configure(m_adapter, &m_adv_handle, &m_adv_data, &m_adv_params);
	}
	check_advert_buffer_for_corruption(advert_data_buffer, "after sd_ble_gap_adv_set_configure");
#endif

	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Failed to set advertisement data. Error %s\n", 
			GetCurrentDateTimeString().c_str(),
			bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		return err_code;
	}
	else
	{
		if (m_iLoggingLevel >= 5)
			printf("Advertising data set successfully\n");
		fflush(stdout);
	}
	
	return NRF_SUCCESS;
}

/**@brief Function for initializing the Advertising functionality and starting advertising.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t advertising_start()
{
	uint32_t             err_code;

	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_ble_gap_adv_stop(m_adapter, m_adv_handle);
	}
	if(err_code != NRF_SUCCESS && err_code != NRF_ERROR_INVALID_STATE)// Often we'll be called when we are not advertising, and get NRF_ERROR_INVALID_STATE
	{
		printf("%s: sd_ble_gap_adv_stop got unexpected error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
	}

#if NRF_SD_BLE_API <= 5
	ble_gap_adv_params_t adv_params;
	memset(&adv_params, 0, sizeof(adv_params));

	adv_params.type = BLE_GAP_ADV_TYPE_ADV_IND;
	adv_params.fp = BLE_GAP_ADV_FP_ANY;
	adv_params.timeout = ADVERTISING_TIMEOUT_3_MIN;
	adv_params.interval = ADVERTISING_INTERVAL_40_MS;
#endif
#if NRF_SD_BLE_API == 2
	adv_params.p_whitelist = NULL;
#endif


	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_ble_gap_adv_start(m_adapter, m_adv_handle, m_config_id);
	}

	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Failed to start advertising. Error %s\n",
			GetCurrentDateTimeString().c_str(),
			bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		return err_code;
	}

	if(m_iLoggingLevel >= 5)
		printf("Started advertising\n");
	fflush(stdout);
	return NRF_SUCCESS;
}

static uint32_t advertising_stop()
{
	uint32_t             err_code;

	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_ble_gap_adv_stop(m_adapter, m_adv_handle);
	}
	if (err_code != NRF_SUCCESS)
	{
		if(err_code != NRF_ERROR_INVALID_STATE)// Often we'll be called when we are not advertising, and get NRF_ERROR_INVALID_STATE
			printf("%s: sd_ble_gap_adv_stop got unexpected error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
	}
	else if (m_iLoggingLevel >= 5)
		printf("Stopped advertising\n");

	fflush(stdout);
	return NRF_SUCCESS;
}


/**@brief Function for handling error message events from sd_rpc.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] code Error code that the error message is associated with.
 * @param[in] message The error message that the callback is associated with.
 */
static void status_handler(adapter_t * adapter, sd_rpc_app_status_t code, const char * message)
{
	printf("%s: Status: %d, message: %s\n", GetCurrentDateTimeString().c_str(), (uint32_t)code, message);
	fflush(stdout);
}

/**@brief Function for handling the log message events from sd_rpc.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] severity Level of severity that the log message is associated with.
 * @param[in] message The log message that the callback is associated with.
 */
static void log_handler(adapter_t * adapter, sd_rpc_log_severity_t severity, const char * message)
{
	switch (severity)
	{
	case SD_RPC_LOG_ERROR:
		printf("%s: Error: %s\n", GetCurrentDateTimeString().c_str(), message);
		fflush(stdout);
		break;

	case SD_RPC_LOG_WARNING:
		printf("%s: Warning: %s\n", GetCurrentDateTimeString().c_str(), message);
		fflush(stdout);
		break;

	case SD_RPC_LOG_INFO:
		printf("%s: Info: %s\n", GetCurrentDateTimeString().c_str(), message);
		fflush(stdout);
		break;
	case SD_RPC_LOG_FATAL:
		printf("%s: Fatal: %s\n", GetCurrentDateTimeString().c_str(), message);
		fflush(stdout);
		break;
	default:
		printf("%s: Log: %s\n", GetCurrentDateTimeString().c_str(), message);
		fflush(stdout);
		break;
	}
}

/**@brief Function for initializing serial communication with the target nRF5 Bluetooth slave.
 *
 * @param[in] serial_port The serial port the target nRF5 device is connected to.
 *
 * @return The new transport adapter.
 */
static adapter_t * adapter_init(char * serial_port, uint32_t baud_rate)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	physical_layer_t  * phy;
	data_link_layer_t * data_link_layer;
	transport_layer_t * transport_layer;

	phy = sd_rpc_physical_layer_create_uart(serial_port,
		baud_rate,
		SD_RPC_FLOW_CONTROL_NONE,
		SD_RPC_PARITY_NONE);
	data_link_layer = sd_rpc_data_link_layer_create_bt_three_wire(phy, 250);
	transport_layer = sd_rpc_transport_layer_create(data_link_layer, 1500);
	return sd_rpc_adapter_create(transport_layer);
}

/**@brief Function for converting a BLE address to a string.
 *
 * @param[in] address       Bluetooth Low Energy address.
 * @param[out] string_buffer The serial port the target nRF5 device is connected to.
 */
static void ble_address_to_string_convert(ble_gap_addr_t address, uint8_t * string_buffer)
{
	const int address_length = 6;
	char      temp_str[3];

	for (int i = address_length - 1; i >= 0; --i)
	{
		sprintf(temp_str, "%02X", address.addr[i]);
		strcat((char *)string_buffer, temp_str);
		if(i > 0)
			strcat((char *)string_buffer, ":");
	}
}

/**
 * @brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  Type of data to be looked for in advertisement data.
 * @param[in]  Advertisement report length and pointer to report.
 * @param[out] If data type requested is found in the data report, type data length and
 *             pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
	uint32_t  index = 0;
	uint8_t * p_data;

	p_data = p_advdata->p_data;

	while (index < p_advdata->data_len)
	{
		uint8_t field_length = p_data[index];
		uint8_t field_type = p_data[index + 1];

		if (field_type == type)
		{
			p_typedata->p_data = &p_data[index + 2];
			p_typedata->data_len = field_length - 1;
			return NRF_SUCCESS;
		}
		index += field_length + 1;
	}
	return NRF_ERROR_NOT_FOUND;
}

/**@brief Function for searching a given name in the advertisement packets.
 *
 * @details Use this function to parse received advertising data and to find a given
 * name in them either as 'complete_local_name' or as 'short_local_name'.
 *
 * @param[in]   p_adv_report   advertising data to parse.
 * @param[in]   name_to_find   name to search.
 * @return   true if the given name was found, false otherwise.
 */
static bool find_adv_name(const ble_gap_evt_adv_report_t *p_adv_report, const char * name_to_find)
{
	uint32_t err_code;
	data_t   adv_data;
	data_t   dev_name;

	// Initialize advertisement report for parsing
#if NRF_SD_BLE_API >= 6
	adv_data.p_data = (uint8_t *)p_adv_report->data.p_data;
	adv_data.data_len = p_adv_report->data.len;
#else
	adv_data.p_data = (uint8_t *)p_adv_report->data;
	adv_data.data_len = p_adv_report->dlen;
#endif

	//search for advertising names
	err_code = adv_report_parse(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME,
		&adv_data,
		&dev_name);
	if (err_code == NRF_SUCCESS)
	{
		if (memcmp(name_to_find, dev_name.p_data, dev_name.data_len) == 0)
		{
			return true;
		}
	}
	else
	{
		// Look for the short local name if it was not found as complete
		err_code = adv_report_parse(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME,
			&adv_data,
			&dev_name);
		if (err_code != NRF_SUCCESS)
		{
			return false;
		}
		if (memcmp(name_to_find, dev_name.p_data, dev_name.data_len) == 0)
		{
			return true;
		}
	}
	return false;
}

/**@brief Function for searching a given UUID in the advertisement packets.
 *
 * @details Use this function to parse received advertising data and to find a given
 * UUID in them as 128-bit UUID
 *
 * @param[in]   p_adv_report   advertising data to parse.
 * @param[in]   name_to_find   name to search.
 * @return   true if the given name was found, false otherwise.
 */
static bool find_adv_uuid(const ble_gap_evt_adv_report_t *p_adv_report, data_t& uuidData)
{
	uint32_t err_code;
	data_t   adv_data;

	// Initialize advertisement report for parsing
#if NRF_SD_BLE_API >= 6
	adv_data.p_data = (uint8_t *)p_adv_report->data.p_data;
	adv_data.data_len = p_adv_report->data.len;
#else
	adv_data.p_data = (uint8_t *)p_adv_report->data;
	adv_data.data_len = p_adv_report->dlen;
#endif

	//search for advertising names
	err_code = adv_report_parse(BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_COMPLETE,
		&adv_data,
		&uuidData);
	if (err_code == NRF_SUCCESS)
	{
		return true;
	}
	return false;
}

/**@brief Function for initializing the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t ble_stack_init()
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	uint32_t            err_code;
	uint32_t *          app_ram_base = NULL;

#if NRF_SD_BLE_API <= 3
	ble_enable_params_t ble_enable_params;
	memset(&ble_enable_params, 0, sizeof(ble_enable_params));
#endif

#if NRF_SD_BLE_API == 3
	ble_enable_params.gatt_enable_params.att_mtu = GATT_MTU_SIZE_DEFAULT;
#elif NRF_SD_BLE_API < 3
	ble_enable_params.gatts_enable_params.attr_tab_size = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
	ble_enable_params.gatts_enable_params.service_changed = false;
	ble_enable_params.common_enable_params.p_conn_bw_counts = NULL;
	ble_enable_params.common_enable_params.vs_uuid_count = 1;
#endif

#if NRF_SD_BLE_API <= 3
	ble_enable_params.gap_enable_params.periph_conn_count = 1;
	ble_enable_params.gap_enable_params.central_conn_count = 1;
	ble_enable_params.gap_enable_params.central_sec_count = 1;

	err_code = sd_ble_enable(m_adapter, &ble_enable_params, app_ram_base);
#else
	err_code = sd_ble_enable(m_adapter, app_ram_base);
#endif

	switch (err_code) {
	case NRF_SUCCESS:
		break;
	case NRF_ERROR_INVALID_STATE:
		printf("BLE stack already enabled\n");
		fflush(stdout);
		break;
	default:
		printf("Failed to enable BLE stack. Error %s\n", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		break;
	}

	return err_code;
}


#if NRF_SD_BLE_API >= 5
/**@brief Function for setting configuration for the BLE stack.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t ble_cfg_set(uint8_t conn_cfg_tag)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	const uint32_t ram_start = 0; // Value is not used by ble-driver
	uint32_t err_code;
	ble_cfg_t ble_cfg;

	// Configure the connection roles.
	memset(&ble_cfg, 0, sizeof(ble_cfg));

#if NRF_SD_BLE_API >= 6
	ble_cfg.gap_cfg.role_count_cfg.adv_set_count = BLE_GAP_ADV_SET_COUNT_DEFAULT;

	ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 1;
	ble_cfg.gap_cfg.role_count_cfg.central_role_count = 1;
	ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 1;

#endif
	ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 0;
	ble_cfg.gap_cfg.role_count_cfg.central_role_count = MAX_CONNECTION_COUNT/*changed from default 1 in original example*/;
	ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 0;

	err_code = sd_ble_cfg_set(m_adapter, BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
	if (err_code != NRF_SUCCESS)
	{
		printf("sd_ble_cfg_set() failed when attempting to set BLE_GAP_CFG_ROLE_COUNT. Error %s\n", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		return err_code;
	}

	memset(&ble_cfg, 0x00, sizeof(ble_cfg));
	ble_cfg.conn_cfg.conn_cfg_tag = conn_cfg_tag;
	ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = 150;

	err_code = sd_ble_cfg_set(m_adapter, BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
	if (err_code != NRF_SUCCESS)
	{
		printf("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GATT. Error %s\n", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		return err_code;
	}

	// Suggestion to do the following to allow multiple connections:
	{
		memset(&ble_cfg, 0x00, sizeof(ble_cfg));
		ble_cfg.conn_cfg.conn_cfg_tag = conn_cfg_tag;// another link said to set this to BLE_CONN_CFG_TAG_CUSTOM;
		ble_cfg.conn_cfg.params.gap_conn_cfg.conn_count = MAX_CONNECTION_COUNT;
		ble_cfg.conn_cfg.params.gap_conn_cfg.event_length = 6/*NRF_SDH_BLE_GAP_EVENT_LENGTH*/;
		err_code = sd_ble_cfg_set(m_adapter, BLE_CONN_CFG_GAP, &ble_cfg, ram_start);
		if (err_code != NRF_SUCCESS)
		{
			printf("sd_ble_cfg_set() failed when attempting to set BLE_CONN_CFG_GAP. Error %s\n", 
				bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
			fflush(stdout);
			return err_code;
		}
	}

	return NRF_SUCCESS;
}
#endif

/**@brief Start scanning (GAP Discovery procedure, Observer Procedure).
 * *
 * @return NRF_SUCCESS on successfully initiating scanning procedure, otherwise an error code.
 */
static uint32_t scan_start()
{
	if(!m_DoScanning)
		return NRF_SUCCESS;
	m_timestampLastScanStart = ::GetTickCount64();
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
#if NRF_SD_BLE_API >= 6
	m_adv_report_buffer.p_data = mp_data;
	m_adv_report_buffer.len = sizeof(mp_data);
#endif

	const ble_gap_scan_params_t m_scan_param = Calls_to_pc_ble_driver::GetScanParams();

	uint32_t err_code = sd_ble_gap_scan_start(m_adapter, &m_scan_param
#if NRF_SD_BLE_API >= 6
		, &m_adv_report_buffer
#endif
	);

	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Scan start failed with error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
	}
	else if(m_iLoggingLevel >= 6)
		printf("scan start success\n");
	fflush(stdout);

	return err_code;
}
static uint32_t scan_restart()
{
	if (!m_DoScanning)
		return NRF_SUCCESS;
	m_timestampLastScanStart = ::GetTickCount64();
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	uint32_t error_code = sd_ble_gap_scan_start(m_adapter, NULL, &m_adv_report_buffer);
	if (error_code == NRF_ERROR_INVALID_STATE)
	{

		const ble_gap_scan_params_t m_scan_param = Calls_to_pc_ble_driver::GetScanParams();

		error_code = sd_ble_gap_scan_start(m_adapter, &m_scan_param
#if NRF_SD_BLE_API >= 6
			, &m_adv_report_buffer
#endif
		);
		if (error_code != NRF_ERROR_INVALID_STATE)
		{
			if (error_code != NRF_SUCCESS)
			{
				printf("%s: re-called sd_ble_gap_scan_start with non-null params after NRF_ERROR_INVALID_STATE, which changed result to different error %s.\n",
					GetCurrentDateTimeString().c_str(),
					bgsdk::Nordic::NordicErrorCodeToMessage(error_code).c_str());
			}
			else if (m_iLoggingLevel >= 7)
			{
				printf("%s: re-called sd_ble_gap_scan_start with non-null params after NRF_ERROR_INVALID_STATE, now successful.\n",
					GetCurrentDateTimeString().c_str());
			}
		}
		else
		{
			printf("%s: re-called sd_ble_gap_scan_start with non-null params after NRF_ERROR_INVALID_STATE, still get NRF_ERROR_INVALID_STATE.\n",
				GetCurrentDateTimeString().c_str());
		}
	}
	else if (error_code != NRF_SUCCESS)
	{
		printf("%s: Scan re-start failed with error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(error_code).c_str());
	}
	else if(m_iLoggingLevel >= 7)
	{
		printf("Scan re-start succeeded\n");
	}
	return error_code;
}
/**@brief Function called upon connecting to BLE peripheral.
 *
 * @details Initiates primary service discovery.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */

void service_discovery_ended()
{
	// Start requesting RSSI values. We'll get event BLE_GAP_EVT_RSSI_CHANGED when RSSI is available.
	// My observation: we only get another event when the RSSI changes.
	// This should keep sending events as long as connection lasts, unless we call sd_ble_gap_rssi_stop.
	if(m_DeviceBeingDiscovered && m_DeviceBeingDiscovered->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
	{
		uint32_t err_code;
		{
			std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic); 
			err_code = sd_ble_gap_rssi_start(m_adapter, m_DeviceBeingDiscovered->m_ConnectionHandle, 10, 0);
		}
		if (err_code != NRF_SUCCESS)
		{
			printf("%s: sd_ble_gap_rssi_start failed error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		}
		else if (m_iLoggingLevel >= 3)
			printf("sd_ble_gap_rssi_start succeeded\n");
	}

	fflush(stdout);

	if(m_iLoggingLevel >= 3)
		printf("service discovery ended, available for new connection\n");
	fflush(stdout);
	
	m_DeviceBeingDiscovered = nullptr;


	scan_start();// re-start scan after service discovery ends
}

static uint32_t service_discovery_start(uint16_t conn_handle)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	// This is the first thing we do after successfully connecting
	uint32_t   err_code;
	uint16_t   start_handle = 0x01;

	if(m_iLoggingLevel >= 3)
		printf("Discovering primary services conn_handle = %d\n", conn_handle);
	fflush(stdout);

	// Initiate procedure to find all services for the device to which we are connected.
	// This causes BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP events to be received (on_service_discovery_response)
	err_code = sd_ble_gattc_primary_services_discover(m_adapter,
		conn_handle, start_handle,
		nullptr);
	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Failed to initiate or continue a GATT Primary Service Discovery procedure. Error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		service_discovery_ended();
	}
	else if(m_iLoggingLevel >= 3)
	{
		printf("sd_ble_gattc_primary_services_discover success for start_handle = 0x%04X\n", start_handle);
		fflush(stdout);
	}

	return err_code;
}
/**@brief Function called upon discovering a BLE peripheral's primary service(s).
 *
 * @details Initiates service's (m_service) characteristic discovery.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t char_discovery_start(uint16_t conn_handle)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	// Characteristic discovery is right after service discovery.
	// This gets the list of characteristics (but not the details).
	if (m_iLoggingLevel >= 3)
		printf("Discovering characteristics conn_handle = %d\n", conn_handle);
	fflush(stdout);

	ble_gattc_handle_range_t next_handle_range;
	next_handle_range.start_handle = m_DeviceBeingDiscovered->m_Services.begin()->m_handle_range.start_handle;
	next_handle_range.end_handle = 0xFFFF;
	// causes event BLE_GATTC_EVT_CHAR_DISC_RSP to get sent (on_characteristic_discovery_response)
	uint32_t err_code = sd_ble_gattc_characteristics_discover(m_adapter, conn_handle, &next_handle_range);
	if(err_code != 0)
	{
		printf("%s: sd_ble_gattc_characteristics_discover error %s, conn_handle = %d start_handle = %d\n",
			GetCurrentDateTimeString().c_str(),
			bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str(),
			conn_handle,
			next_handle_range.start_handle);
		fflush(stdout);
	}
	else if (m_iLoggingLevel >= 3)
	{
		printf("sd_ble_gattc_characteristics_discover success for start_handle 0x%04X\n", next_handle_range.start_handle);
		fflush(stdout);
	}
	return err_code;
}
void characteristic_metadata_discovery_start()
{
	// This fills in two kinds of details:
	//		Look for any unknown UUIDs that were noticed in on_service_discovery_response
	//		Get the metadata of every characteristic
	// This article tells how to find the full 128-bit UUID for services: https://devzone.nordicsemi.com/f/nordic-q-a/17946/how-to-find-128-bit-uuid-services-with-sd_ble_gattc_primary_services_discover
	// Briefy: call sd_ble_gattc_read()

	for(size_t nServices = m_DeviceBeingDiscovered->m_Services.size(), iService = 0; iService < nServices; ++iService )
	{
		if (m_DeviceBeingDiscovered->m_ConnectionHandle == BLE_CONN_HANDLE_INVALID)// check, in case we got a disconnect
		{
			service_discovery_ended();
			return;
		}
		if(m_DeviceBeingDiscovered->m_Services[iService].m_uuid_type == BLE_UUID_TYPE_UNKNOWN &&
			m_DeviceBeingDiscovered->m_Services[iService].m_uuid_string.empty())
		{
			// This will trigger event BLE_GATTC_EVT_READ_RSP (on_gattc_evt_read_response)
			uint32_t err_code;
			{
				std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic); 
				err_code = sd_ble_gattc_read(m_adapter, m_DeviceBeingDiscovered->m_ConnectionHandle, m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle, 0/*offset*/);
			}
			if(err_code != NRF_SUCCESS)
			{
				printf("%s: sd_ble_gattc_read failed during characteristic_metadata_discovery_start error %s, conn_handle = %d\n", 
					GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str(), m_DeviceBeingDiscovered->m_ConnectionHandle);
				service_discovery_ended();
			}
			else if(m_iLoggingLevel >= 3)
			{
				printf("sd_ble_gattc_read success to find 128-bit UUID of service 0x%04X, conn_handle = %d\n",
					m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle,
					m_DeviceBeingDiscovered->m_ConnectionHandle);
			}
			fflush(stdout);
			return;// when we get the read event, we'll continue discovery
		}
	}
	if (m_DeviceBeingDiscovered->m_ConnectionHandle == BLE_CONN_HANDLE_INVALID)// check, in case we got a disconnect
	{
		service_discovery_ended();
		return;
	}
	// If we get here, we have all the 128-bit UUIDs for services.
	// Now get metadata of all characteristics
	for(auto serv : m_DeviceBeingDiscovered->m_Services)
	{
		for(auto ch : serv.m_Characteristics)
		{
			if(ch.m_bGotDescriptionDetails == false)
			{
				// This will trigger event BLE_GATTC_EVT_READ_RSP (on_gattc_evt_read_response)
				uint32_t err_code;
				{
					std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic); 
					err_code = sd_ble_gattc_read(
					m_adapter,
					m_DeviceBeingDiscovered->m_ConnectionHandle,
					ch.m_DescriptionHandle, 0/*offset*/);
				}
				if (err_code != NRF_SUCCESS)
				{
					printf("%s: sd_ble_gattc_read failed during characteristic_metadata_discovery_start error %s, conn_handle = %d\n", 
						GetCurrentDateTimeString().c_str(),
						bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str(), m_DeviceBeingDiscovered->m_ConnectionHandle);
					fflush(stdout);
					service_discovery_ended();
				}
				else if (m_iLoggingLevel >= 3)
				{
					printf("sd_ble_gattc_read started for characteristic description handle 0x%04X, conn_handle = %d\n",
						ch.m_DescriptionHandle,
						m_DeviceBeingDiscovered->m_ConnectionHandle);
					fflush(stdout);
				}
				return;// when we get the read event, we'll continue discovery
			}
		}
	}
	// If we get here, we are entirely done getting metadata for a phone
	if(m_iLoggingLevel >= 3)
		printf("entirely done getting metadata from phone\n");
	service_discovery_ended();
}

void on_gattc_evt_read_response(const ble_gattc_evt_t * const p_ble_gattc_evt)// event BLE_GATTC_EVT_READ_RSP
{
	::SetEvent(hEvent_BLE_GATTC_EVT_READ_RSP);
	if(m_DeviceBeingDiscovered == nullptr)
	{
		// when we are not discovering services, the only characteristic we read is in the keepalive thread,
		// which reads a single byte bitmask from the phone
		if(m_iLoggingLevel >= 5)
			printf("%s: keepalive conn_handle = %d characteristic handle 0x%04X #bytes = %d value %c\n",
				GetCurrentDateTimeString().c_str(),
				p_ble_gattc_evt->conn_handle,
				p_ble_gattc_evt->params.read_rsp.handle,
				p_ble_gattc_evt->params.read_rsp.len,
				(p_ble_gattc_evt->params.read_rsp.len == 1 ? (p_ble_gattc_evt->params.read_rsp.data[0]+'0') : ' '));
		fflush(stdout);
		return;
	}

	if(p_ble_gattc_evt->gatt_status != BLE_GATT_STATUS_SUCCESS)
	{
		printf("%s: on_gattc_evt_read_response conn_handle = %d error 0x%04X\n", GetCurrentDateTimeString().c_str(), p_ble_gattc_evt->conn_handle, p_ble_gattc_evt->gatt_status); fflush(stdout);
		service_discovery_ended();
		return;
	}
	else
	{
		// Fill in UUID in service list
		for (size_t nServices = m_DeviceBeingDiscovered->m_Services.size(), iService = 0; iService < nServices; ++iService)
		{
			if (m_DeviceBeingDiscovered->m_Services[iService].m_uuid_type == BLE_UUID_TYPE_UNKNOWN
				&& m_DeviceBeingDiscovered->m_Services[iService].m_uuid_string.empty()
				&& m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle == p_ble_gattc_evt->params.read_rsp.handle)
			{

				// Currently the only characteristic we read is 128-bit UUID values (16 bytes)
				if (p_ble_gattc_evt->params.read_rsp.len == 16)
				{
					char strUUID[UUID_CHARACTER_LENGTH_NONULL + 1/*null*/];
					if (UUID_to_string_convert(p_ble_gattc_evt->params.read_rsp.len, p_ble_gattc_evt->params.read_rsp.data, strUUID, sizeof(strUUID) / sizeof(strUUID[0])))
					{
						if (m_iLoggingLevel >= 3)
							printf("found service 0x%04X UUID = %s\n",
								m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle,
								strUUID);
						fflush(stdout);
						m_DeviceBeingDiscovered->m_Services[iService].m_uuid_string = strUUID;
						// kick off finding the next bit of characteristic metadata
						characteristic_metadata_discovery_start();
						return;
					}
				}
				else
				{
					printf("%s: on_gattc_evt_read_response conn_handle = %d has unexpected length\n", GetCurrentDateTimeString().c_str(), p_ble_gattc_evt->conn_handle); fflush(stdout);
				}
			}
		}

		// See if the handle we read matches any characteristic metadata
		for (auto& serv : m_DeviceBeingDiscovered->m_Services)
		{
			for (auto& ch : serv.m_Characteristics)
			{
				if (ch.m_DescriptionHandle == p_ble_gattc_evt->params.read_rsp.handle)
				{
					ch.m_bGotDescriptionDetails = true;

					{
						std::string hexData;
						hexData.resize(p_ble_gattc_evt->params.read_rsp.len*3, ' ');
						for(int iByte = 0; iByte < p_ble_gattc_evt->params.read_rsp.len; ++iByte)
						{
							char thisByte[3];
							sprintf(thisByte, "%02x", p_ble_gattc_evt->params.read_rsp.data[iByte]);
							hexData[iByte * 3 + 0] = thisByte[0];
							hexData[iByte * 3 + 1] = thisByte[1];
							hexData[iByte * 3 + 2];
						}
						if (m_iLoggingLevel >= 3)
							printf("MATCHED on_gattc_evt_read_response conn_handle = %d read handle = 0x%04X #bytes = %d offset = %d data = %s\n",
								p_ble_gattc_evt->conn_handle,
								p_ble_gattc_evt->params.read_rsp.handle,
								p_ble_gattc_evt->params.read_rsp.len,
								p_ble_gattc_evt->params.read_rsp.offset,
								hexData.c_str());
						fflush(stdout);
						if(p_ble_gattc_evt->params.read_rsp.len == 19)// has the UUID of a characteristic
						{
							uint16_t valueCharacteristicHandle = p_ble_gattc_evt->params.read_rsp.data[1] + (p_ble_gattc_evt->params.read_rsp.data[2] << 8);
							char strUUID[UUID_CHARACTER_LENGTH_NONULL + 1/*null*/];
							if (UUID_to_string_convert(16, &p_ble_gattc_evt->params.read_rsp.data[3], strUUID, sizeof(strUUID) / sizeof(strUUID[0])))
							{
								if(ch.m_ValueHandle == valueCharacteristicHandle)
								{
									ch.m_ValueHandleUUID = strUUID;
									if(m_iLoggingLevel >= 3)
										printf("FOUND characteristic handle 0x%04X UUID = %s\n",
											ch.m_ValueHandle,
											strUUID);
								}
								else
								{
									printf("%s: UNEXPECTED received value handle 0x%04X != expected value handle 0x%04X\n",
										GetCurrentDateTimeString().c_str(),
										valueCharacteristicHandle,
										ch.m_ValueHandle);
								}
								fflush(stdout);
							}
						}

					}

					// kick off finding the next bit of characteristic metadata
					characteristic_metadata_discovery_start();
					return;
				}
			}
		}

		printf("%s: on_gattc_evt_read_response conn_handle = %d read handle = 0x%04X #bytes = %d offset = %d was not for any characteristic we are discovering\n",
			GetCurrentDateTimeString().c_str(),
			p_ble_gattc_evt->conn_handle,
			p_ble_gattc_evt->params.read_rsp.handle,
			p_ble_gattc_evt->params.read_rsp.len,
			p_ble_gattc_evt->params.read_rsp.offset);
		fflush(stdout);
	}
}


/**@brief Function that write's the HRM characteristic's CCCD.
 * *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t hrm_cccd_set(uint16_t conn_handle, uint8_t value)
{
	ble_gattc_write_params_t write_params;
	uint8_t                  cccd_value[2] = { value, 0 };

	printf("Setting HRM CCCD\n");
	fflush(stdout);

	if (m_hrm_cccd_handle == 0)
	{
		printf("Error. No CCCD handle has been found\n");
		fflush(stdout);
		return NRF_ERROR_INVALID_STATE;
	}

	write_params.handle = m_hrm_cccd_handle;
	write_params.len = 2;
	write_params.p_value = cccd_value;
	write_params.write_op = BLE_GATT_OP_WRITE_REQ;
	write_params.offset = 0;

	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	return sd_ble_gattc_write(m_adapter, conn_handle, &write_params);
}


/** Event functions */

/**@brief Function called on BLE_GAP_EVT_CONNECTED event.
 *
 * @details Update connection state and proceed to discovering the peer's GATT services.
 *
 * @param[in] p_ble_gap_evt GAP event.
 */
static const char* Nordic_ble_gap_role_to_string(uint8_t role)
{
	switch(role)
	{
	case BLE_GAP_ROLE_INVALID:
		return "invalid";
	case BLE_GAP_ROLE_PERIPH:
		return "peripheral";
	case BLE_GAP_ROLE_CENTRAL:
		return "central";
	default:
		return "(programming error)";
	}
}
static void on_connected(const uint16_t conn, const char* pRoleDescription)
{
	if (m_iLoggingLevel >= 2)
		printf("%s: Thread %d: received BLE_GAP_EVT_CONNECTED conn_handle = %d role = %s\n", 
		GetCurrentDateTimeString().c_str(),
		::GetCurrentThreadId(),
		conn,
		pRoleDescription);
	fflush(stdout);

	m_DeviceBeingDiscovered->m_ConnectionHandle = conn;

	service_discovery_start(conn);
}

/**@brief Function called on BLE_GAP_EVT_ADV_REPORT event.
 *
 * @details Create a connection if received advertising packet corresponds to desired BLE device.
 *
 * @param[in] p_ble_gap_evt Advertising Report Event.
 */
static void on_adv_report(const ble_gap_addr_t peerAddr, const std::string strUUID, int8_t rssi)
{
	m_timestampLastAdvertisement = ::GetTickCount64();
	uint32_t err_code;
	uint8_t  str[STRING_BUFFER_SIZE] = { 0 };

	// Log the Bluetooth device address of advertisement packet received.
	ble_address_to_string_convert(peerAddr, str);

	fflush(stdout);

	{
		m_all128bitUUIDDevices.insert(strUUID);
		++m_totalAdvertisementsReceivedWith128BitUUIDs;
		auto connDev = ConnectedDevices.find(strUUID);
		if(!m_DeviceBeingDiscovered && connDev != ConnectedDevices.end() && connDev->second->m_ConnectionHandle == BLE_CONN_HANDLE_INVALID)
		{
			bool bConnectSuccess = false;// set to true if all goes well
			::ResetEvent(hEvent_BLE_GAP_EVT_CONNECT_Completed);
			current_CONNECT_timed_out = false;

			m_DeviceBeingDiscovered = connDev->second;
			m_DeviceBeingDiscovered->m_numberRSSIEvents = 0;// restart whenever we re-connect
			if(m_iLoggingLevel >= 1)
				printf("%s: Thread %d: FOUND phone with UUID %s RSSI = %d, MAC = %s\n", 
				GetCurrentDateTimeString().c_str(),
				::GetCurrentThreadId(),
				strUUID.c_str(), rssi, str);
			{

				const ble_gap_scan_params_t m_scan_param =
				{
				#if NRF_SD_BLE_API >= 6
					0,                       // Set if accept extended advertising packetets.
					0,                       // Set if report inomplete reports.
				#endif
					0,                       // Set if active scanning.
				#if NRF_SD_BLE_API < 6
					0,                       // Set if selective scanning.
				#endif
				#if NRF_SD_BLE_API >= 6
					BLE_GAP_SCAN_FP_ACCEPT_ALL,
					BLE_GAP_PHY_1MBPS,
				#endif
				#if NRF_SD_BLE_API == 2
					NULL,                    // Set white-list.
				#endif
				#if NRF_SD_BLE_API == 3 || NRF_SD_BLE_API == 5
					0,                       // Set adv_dir_report.
				#endif
					(uint16_t)SCAN_INTERVAL,
					(uint16_t)SCAN_WINDOW,
					(uint16_t)SCAN_TIMEOUT
				#if NRF_SD_BLE_API >= 6
					, { 0 }                  // Set chennel mask.
				#endif
				};
				std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
				err_code = sd_ble_gap_connect(m_adapter,
					&(peerAddr),
					&m_scan_param,
					&m_connection_param
		#if NRF_SD_BLE_API >= 5
					, m_config_id
		#endif
				);
			}
			if (err_code != NRF_SUCCESS)
			{
				if(err_code == NRF_ERROR_CONN_COUNT)
					printf("%s: Connection Request Failed because connection count exceeded\n", GetCurrentDateTimeString().c_str());
				else
					printf("%s: Connection Request Failed error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
				fflush(stdout);
				service_discovery_ended();
			}
			else
			{
				if (m_iLoggingLevel >= 2)
					printf("%s: Thread %d: connection request succeeded, connection in progress\n",
					GetCurrentDateTimeString().c_str(),
					::GetCurrentThreadId());
				fflush(stdout);
				// We'll get BLE_GAP_EVT_CONNECTED if the connection request succeeds
				DWORD millisecondsToWaitForConnectCompletion = 65000;
				DWORD dWait = ::WaitForSingleObject(hEvent_BLE_GAP_EVT_CONNECT_Completed, millisecondsToWaitForConnectCompletion);
				if (dWait == WAIT_OBJECT_0)
				{
					if (current_CONNECT_timed_out)
					{
						printf("%s: sd_ble_gap_connect got BLE_GAP_TIMEOUT_SRC_CONN (which is allowable), end this connection attempt\n", GetCurrentDateTimeString().c_str());
						// Note: no need to call sd_ble_gap_connect_cancel since we got the timeout event
					}
					else
					{
						bConnectSuccess = true;
						if (m_iLoggingLevel >= 2)
						{
							printf("%s: Thread %d: got expected BLE_GAP_EVT_CONNECTED\n",
								GetCurrentDateTimeString().c_str(),
								::GetCurrentThreadId()
							);
						}
					}
				}
				else if (dWait == WAIT_TIMEOUT)
				{
					printf("%s: Thread %d: ERROR: never got BLE_GAP_EVT_CONNECTED or BLE_GAP_TIMEOUT_SRC_CONN after waiting %d milliseconds.\n",
						GetCurrentDateTimeString().c_str(),
						::GetCurrentThreadId(),
						millisecondsToWaitForConnectCompletion
					);
					err_code = sd_ble_gap_connect_cancel(m_adapter);
					if (err_code != NRF_SUCCESS)
						printf("%s: sd_ble_gap_connect_cancel failed %s", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());

					printf("Exit test program immediately\n");
					::ExitProcess(0/*exit code*/);
				}
				else
				{
					printf("ERROR: WaitForSingleObject returned unexpected value %d while waiting for BLE_GATTC_EVT_READ_RSP\n", dWait);
				}
			}

			if (bConnectSuccess == false)
			{
				service_discovery_ended();
			}
			else
				service_discovery_start(m_DeviceBeingDiscovered->m_ConnectionHandle);
		}
		else if(m_iLoggingLevel >= 6)
		{
			printf("Advertisement RSSI = %d, MAC = %s, UUID = %s\n", rssi, str, strUUID.c_str());
			fflush(stdout);
		}
	}
}

/**@brief Function called on BLE_GAP_EVT_TIMEOUT event.
 *
 * @param[in] ble_gap_evt_t Timeout Event.
 */
static void on_ble_gap_evt_timeout(const ble_gap_evt_t * const p_ble_gap_evt)
{
	if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
	{
		if (m_DeviceBeingDiscovered)
		{
			printf("%s: BLE_GAP_TIMEOUT_SRC_CONN timeout for connection\n", GetCurrentDateTimeString().c_str());
			current_CONNECT_timed_out = true;
			::SetEvent(hEvent_BLE_GAP_EVT_CONNECT_Completed);
		}
		else
			printf("%s: ERROR: got timeout event BLE_GAP_TIMEOUT_SRC_CONN when we are not attempting to connect\n", GetCurrentDateTimeString().c_str());

	}
	else if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
	{
		scan_start();//re-start scan after scan timeout, which should never happen with our scan parameters
	}
}

static void on_gattc_timeout()// called for event BLE_GATTC_EVT_TIMEOUT
{
	// Let GATTC read/write operations know they are done, with a timeout
	if (hEvent_current_GATTC)
	{
		printf("%s: current GATTC read/write timed out\n",
				GetCurrentDateTimeString().c_str());
		current_GATTC_timed_out = true;
		::SetEvent(hEvent_current_GATTC);
	}
	else
		printf("%s: received BLE_GATTC_EVT_TIMEOUT when not doing a GATTC read/write\n", GetCurrentDateTimeString().c_str());
}

/**@brief Function called on BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP event.
 *
 * @details Update service state and proceed to discovering the service's GATT characteristics.
 *
 * @param[in] p_ble_gattc_evt Primary Service Discovery Response Event.
 */
static void on_service_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	int count;
	const ble_gattc_service_t * service;

	if(p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_INVALID_PDU // often get this when we reach the end 
		|| p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_INVALID_HANDLE // sometimes get this when we reach the end
		|| p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND // on iPhone 7, final error is BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND
		)
	{
		// All services have been discovered.
		if(m_iLoggingLevel >= 3)
			printf("finished finding services, conn_handle = %d\n", p_ble_gattc_evt->conn_handle);
		fflush(stdout);
		// Start getting meta-data about all characteristics of all services
		char_discovery_start(p_ble_gattc_evt->conn_handle);
		return;
	}
	else if (p_ble_gattc_evt->gatt_status != BLE_GATT_STATUS_SUCCESS)
	{
		printf("%s: Service discovery failed. Error code 0x%X %s error handle 0x%04X\n", 
			GetCurrentDateTimeString().c_str(),
			p_ble_gattc_evt->gatt_status,
			((p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND) ? "BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND" :
			(p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_INVALID_HANDLE) ? "BLE_GATT_STATUS_ATTERR_INVALID_HANDLE" : ""),
			p_ble_gattc_evt->error_handle
		); 
		fflush(stdout);
		service_discovery_ended();
		return;
	}

	count = p_ble_gattc_evt->params.prim_srvc_disc_rsp.count;

	if (count == 0)
	{
		printf("%s: Service discovery has zero count\n", GetCurrentDateTimeString().c_str());
		fflush(stdout);
		service_discovery_ended();
		return;
	}

	uint16_t last_end_handle = 0;
	for(int service_index = 0; service_index < count; ++service_index)
	{
		service = &(p_ble_gattc_evt->params.prim_srvc_disc_rsp.services[service_index]);

		switch(service->uuid.type)
		{
		case BLE_UUID_TYPE_UNKNOWN: /**< Invalid UUID type. this means it is a 128-bit UUID that we must read separately (later) */
			if (m_iLoggingLevel >= 3)
				printf("service type 128-bit UUID: handle_range start = 0x%04X end = 0x%04X conn_handle = %d\n",
					service->handle_range.start_handle,
					service->handle_range.end_handle,
					p_ble_gattc_evt->conn_handle);
			fflush(stdout);
			m_DeviceBeingDiscovered->m_Services.push_back(ConnectedService(service->uuid.type, service->handle_range));
			break;
		case BLE_UUID_TYPE_BLE: /**< Bluetooth SIG UUID (16-bit). */
			if (m_iLoggingLevel >= 3)
				printf("service type 16-bit, UUID = 0x%04X, handle_range start = 0x%04X end = 0x%04X conn_handle = %d\n", 
					service->uuid.uuid, 
					service->handle_range.start_handle,
					service->handle_range.end_handle, 
					p_ble_gattc_evt->conn_handle);
			fflush(stdout);
			m_DeviceBeingDiscovered->m_Services.push_back(ConnectedService(service->uuid.type, service->uuid.uuid, service->handle_range));
			break;
		case BLE_UUID_TYPE_VENDOR_BEGIN: // 128-bit
			if (m_iLoggingLevel >= 1)
				printf("%s: service type 128-bit, octets 12-13 = 0x%04X (not handled by test program), handle_range start = 0x%04X end = 0x%04X conn_handle = %d\n",
					GetCurrentDateTimeString().c_str(),
					service->uuid.uuid,
					service->handle_range.start_handle,
					service->handle_range.end_handle, 
					p_ble_gattc_evt->conn_handle);
			fflush(stdout);
			break;
		default:
			printf("%s: Unknown service type discovered: 0x%04X conn_handle = %d\n", 
				GetCurrentDateTimeString().c_str(),
				service->uuid.type,
				p_ble_gattc_evt->conn_handle);
			return;
		}
		last_end_handle = service->handle_range.end_handle;
	}
	fflush(stdout);

	// Continue to find more services
	uint32_t   err_code;
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_ble_gattc_primary_services_discover(m_adapter,
		p_ble_gattc_evt->conn_handle, last_end_handle+1,
		nullptr);
	}
	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Failed to continue a GATT Primary Service Discovery procedure. Error %s\n", GetCurrentDateTimeString().c_str(),  bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		service_discovery_ended();
	}
	else if(m_iLoggingLevel >= 3)
	{
		printf("sd_ble_gattc_primary_services_discover success for handle 0x%04X\n", last_end_handle + 1);
		fflush(stdout);
	}
}

/**@brief Function called on BLE_GATTC_EVT_CHAR_DISC_RSP event.
 *
 * @details Update characteristic state and proceed to discovering the characteristicss descriptors.
 *
 * @param[in] p_ble_gattc_evt Characteristic Discovery Response Event.
 */
static void on_characteristic_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	// This tells us the set of characteristics for services on the phone (but not the details of the characteristic metadata)

	int count = p_ble_gattc_evt->params.char_disc_rsp.count;

	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		if(p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND)
		{
			if (m_iLoggingLevel >= 3)
				printf("Characteristic discovery done conn_handle = %d\n", p_ble_gattc_evt->conn_handle);
		}
		else
			printf("%s: Characteristic discovery failed. Error code 0x%X, conn_handle = %d\n", GetCurrentDateTimeString().c_str(), p_ble_gattc_evt->gatt_status, p_ble_gattc_evt->conn_handle);
		fflush(stdout);
		// Now that we have the list of characteristics, get metadata details
		characteristic_metadata_discovery_start();
		return;
	}
	else if(count == 0)
	{
		printf("%s: Characteristic discovery has 0 characteristics (unexpected!), conn_handle = %d\n", GetCurrentDateTimeString().c_str(), p_ble_gattc_evt->conn_handle);
		// Now that we have the list of characteristics, get metadata details
		characteristic_metadata_discovery_start();
		return;
	}

	for (int i = 0; i < count; i++)
	{
		uint16_t desc_handle = p_ble_gattc_evt->params.char_disc_rsp.chars[i].handle_decl;
		uint16_t val_handle = p_ble_gattc_evt->params.char_disc_rsp.chars[i].handle_value;

		// Find the service for this characteristic
		bool bFoundService = false;
		for(size_t nServices = m_DeviceBeingDiscovered->m_Services.size(), iService = 0; iService < nServices; ++iService)
		{
			if(	   desc_handle >= m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle
				&& desc_handle <= m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.end_handle
				&& val_handle >= m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle
				&& val_handle <= m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.end_handle)
			{
				bFoundService = true;
				if (m_iLoggingLevel >= 3)
					printf("Service 0x%04X Characteristic desc 0x%04X, value 0x%04X UUID type %d, short UUID: 0x%04X, conn_handle = %d\n",
						m_DeviceBeingDiscovered->m_Services[iService].m_handle_range.start_handle,
						desc_handle,
						val_handle,
						p_ble_gattc_evt->params.char_disc_rsp.chars[i].uuid.type,
						p_ble_gattc_evt->params.char_disc_rsp.chars[i].uuid.uuid,
						p_ble_gattc_evt->conn_handle);
				fflush(stdout);
				m_DeviceBeingDiscovered->m_Services[iService].m_Characteristics.push_back(ConnectedCharacteristic(desc_handle, val_handle));
			}
		}
		if(bFoundService == false)
		{
			printf("%s: UNFOUND Characteristic desc handle 0x%04X, value handle 0x%04X UUID type %d, short UUID: 0x%04X, conn_handle = %d\n",
				GetCurrentDateTimeString().c_str(),
				desc_handle,
				val_handle,
				p_ble_gattc_evt->params.char_disc_rsp.chars[i].uuid.type,
				p_ble_gattc_evt->params.char_disc_rsp.chars[i].uuid.uuid,
				p_ble_gattc_evt->conn_handle);
			fflush(stdout);
		}
	}
	// ask for more characteristic meta-data
	// causes event BLE_GATTC_EVT_CHAR_DISC_RSP to get sent (on_characteristic_discovery_response)
	ble_gattc_handle_range_t next_handle_range;
	next_handle_range.start_handle = p_ble_gattc_evt->params.char_disc_rsp.chars[count-1].handle_decl+1;
	next_handle_range.end_handle = 0xFFFF;
	uint32_t err_code;
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic); 
		err_code = sd_ble_gattc_characteristics_discover(m_adapter, p_ble_gattc_evt->conn_handle, &next_handle_range);
	}
	if(err_code != 0)
	{
		printf("%s: sd_ble_gattc_characteristics_discover error %s, conn_handle = %d for start_handle = %d\n", 
			GetCurrentDateTimeString().c_str(),
			bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str(),
			p_ble_gattc_evt->conn_handle,
			next_handle_range.start_handle);
		service_discovery_ended();
	}
	else if (m_iLoggingLevel >= 3)
	{
		printf("sd_ble_gattc_characteristics_discover success for start_handle 0x%04X\n", next_handle_range.start_handle);
		fflush(stdout);
	}
}

/**@brief Function called on BLE_GATTC_EVT_DESC_DISC_RSP event.
 *
 * @details Update CCCD descriptor state and proceed to prompting user to toggle notifications.
 *
 * @param[in] p_ble_gattc_evt Descriptor Discovery Response Event.
 */
static void on_descriptor_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	int count = p_ble_gattc_evt->params.desc_disc_rsp.count;

	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		printf("%s: Descriptor discovery failed. Error code 0x%X\n", GetCurrentDateTimeString().c_str(), p_ble_gattc_evt->gatt_status);
		fflush(stdout);
		return;
	}

	printf("%s: Received descriptor discovery response, descriptor count: %d\n", GetCurrentDateTimeString().c_str(), count);
	fflush(stdout);

	for (int i = 0; i < count; i++)
	{
		printf("Descriptor handle: 0x%04X, UUID: 0x%04X\n",
			p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle,
			p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid);
		fflush(stdout);

		if (p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_CCCD)
		{
			m_hrm_cccd_handle = p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle;
			printf("Press enter to toggle notifications on the HRM characteristic\n");
			fflush(stdout);
		}
	}
}

/**@brief Function called on BLE_GATTC_EVT_WRITE_RSP event.
 *
 * @param[in] p_ble_gattc_evt Write Response Event.
 */
static void on_write_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		printf("%s: Error. Write operation failed. Error code 0x%X %s\n",
			GetCurrentDateTimeString().c_str(),
			p_ble_gattc_evt->gatt_status,
			(p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_WRITE_NOT_PERMITTED ? "BLE_GATT_STATUS_ATTERR_WRITE_NOT_PERMITTED" : ""));
	}
	else if (m_iLoggingLevel >= 4)
		printf("Received event BLE_GATTC_EVT_WRITE_RSP with successfull GATT status\n");
	::SetEvent(hEvent_BLE_GATTC_EVT_WRITE_RSP);
}

/**@brief Function called on BLE_GATTC_EVT_HVX event.
 *
 * @details Logs the received heart rate measurement.
 *
 * @param[in] p_ble_gattc_evt Handle Value Notification/Indication Event.
 */
static void on_hvx(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	if (m_iLoggingLevel >= 4)
		printf("Received data for characteristic handle 0x%04X #bytes = %d hvx handle = 0x%04X\n", 
		p_ble_gattc_evt->params.hvx.handle,
		p_ble_gattc_evt->params.hvx.len,
		p_ble_gattc_evt->params.hvx.handle);
	//p_ble_gattc_evt->params.hvx.data[1]);
	fflush(stdout);
}

/**@brief Function called on BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST event.
 *
 * @details Update GAP connection parameters.
 *
 * @param[in] p_ble_gap_evt Connection Parameter Update Event.
 */
static void on_conn_params_update_request(const ble_gap_evt_t * const p_ble_gap_evt)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	printf("%s: on_conn_params_update_request conn_handle = %d\n", GetCurrentDateTimeString().c_str(), p_ble_gap_evt->conn_handle);fflush(stdout);
	uint32_t err_code = sd_ble_gap_conn_param_update(m_adapter, p_ble_gap_evt->conn_handle,
		&(p_ble_gap_evt->
			params.conn_param_update_request.conn_params));
	if (err_code != NRF_SUCCESS)
	{
		printf("%s: Conn params update failed, error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
	}
}

#if NRF_SD_BLE_API >= 3
/**@brief Function called on BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST event.
 *
 * @details Replies to an ATT_MTU exchange request by sending an Exchange MTU Response to the client.
 *
 * @param[in] p_ble_gatts_evt Exchange MTU Request Event.
 */
static void on_exchange_mtu_request(const ble_gatts_evt_t * const p_ble_gatts_evt)
{
	std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
	printf("%s: on_exchange_mtu_request\n", GetCurrentDateTimeString().c_str());fflush(stdout);
	uint32_t err_code = sd_ble_gatts_exchange_mtu_reply(
		m_adapter,
		p_ble_gatts_evt->conn_handle,
#if NRF_SD_BLE_API < 5
		GATT_MTU_SIZE_DEFAULT);
#else
		BLE_GATT_ATT_MTU_DEFAULT);
#endif

	if (err_code != NRF_SUCCESS)
	{
		printf("%s: MTU exchange request reply failed, error %s\n", GetCurrentDateTimeString().c_str(), bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
	}
}

/**@brief Function called on BLE_GATTC_EVT_EXCHANGE_MTU_RSP event.
 *
 * @details Logs the new BLE server RX MTU size.
 *
 * @param[in] p_ble_gattc_evt Exchange MTU Response Event.
 */
static void on_exchange_mtu_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	uint16_t server_rx_mtu = p_ble_gattc_evt->params.exchange_mtu_rsp.server_rx_mtu;

	printf("%s: MTU response received. New ATT_MTU is %d\n", GetCurrentDateTimeString().c_str(), server_rx_mtu);
	fflush(stdout);
}
#endif


/** Event dispatcher */

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in] adapter The transport adapter.
 * @param[in] p_ble_evt Bluetooth stack event.
 */
static void ble_evt_dispatch(adapter_t * adapter, ble_evt_t * p_ble_evt)
{
	if(m_bShuttingDown)
		return;

	static bool bFirstTime = true;
	if (bFirstTime)
	{
		printf("%s: Nordic events are received on thread %d\n", GetCurrentDateTimeString().c_str(), ::GetCurrentThreadId()); fflush(stdout);
		bFirstTime = false;
	}

	if (p_ble_evt == NULL)
	{
		printf("%s: Received an empty BLE event\n", GetCurrentDateTimeString().c_str());
		fflush(stdout);
		return;
	}

	switch (p_ble_evt->header.evt_id)
	{
	case BLE_GAP_EVT_CONNECTED:
		if (m_DeviceBeingDiscovered)
		{
			if (m_iLoggingLevel >= 2)
				printf("%s: Got BLE_GAP_EVT_CONNECTED %s\n", GetCurrentDateTimeString().c_str(), Nordic_ble_gap_role_to_string(p_ble_evt->evt.gap_evt.params.connected.role));
			current_CONNECT_timed_out = false;
			m_DeviceBeingDiscovered->m_ConnectionHandle = p_ble_evt->evt.gap_evt.conn_handle;
			::SetEvent(hEvent_BLE_GAP_EVT_CONNECT_Completed);
		}
		else
			printf("%s: ERROR: got connection event BLE_GAP_EVT_CONNECTED when we are not attempting to connect\n", GetCurrentDateTimeString().c_str());
		break;

	case BLE_GAP_EVT_DISCONNECTED:
	{
		if(m_iLoggingLevel >= 1)
			printf("%s: event handler thread got BLE_GAP_EVT_DISCONNECTED, post to worker thread to handle\n", GetCurrentDateTimeString().c_str());

		uint8_t reason = p_ble_evt->evt.gap_evt.params.disconnected.reason;
		uint16_t conn = p_ble_evt->evt.gap_evt.conn_handle;
		// 'remove_connected_device' will mark the member of ConnectedDevices for this connection as BLE_CONN_HANDLE_INVALID.
		// Since m_DeviceBeingDiscovered is an element of ConnectedDevices, it thereby gets changed to BLE_CONN_HANDLE_INVALID
		remove_connected_device(conn);
		// If there is a read or write in progress, let it know that it should give up
		if (hEvent_current_GATTC)
		{
			printf("%s: BLE_GAP_EVT_DISCONNECTED while GATTC read/write in progress. Cancel read/write.\n", GetCurrentDateTimeString().c_str());
			current_GATTC_timed_out = true;
			::SetEvent(hEvent_current_GATTC);
		}
		// If we initiated the disconnect by calling sd_ble_gap_disconnect, then a thread is waiting on this event
		::SetEvent(hEvent_BLE_GAP_EVT_DISCONNECTED);
		// We spin off a handler in our worker thread to handle unexpected disconnects
		message_queue_.Post([reason, conn]() {
			try {
				if (m_iLoggingLevel >= 1)
					printf("%s: Thread %d: BLE_GAP_EVT_DISCONNECTED received, reason: %s, conn_handle = %d\n",
					GetCurrentDateTimeString().c_str(),
					::GetCurrentThreadId(),
					bgsdk::Nordic::Nordic_BLE_HCI_STATUS_CODES_ToMessage(reason).c_str(),
					conn);

				if (m_DeviceBeingDiscovered)
				{
					m_DeviceBeingDiscovered->m_ConnectionHandle = BLE_CONN_HANDLE_INVALID;
					service_discovery_ended();
				}
			}
			catch (const std::exception& ex)
			{
				printf("caught exception from disconnect handler: %s\n", ex.what());
			}
			catch (...)
			{
				printf("caught non-std::exception from disconnect handler\n");
			}
		});
	}
		break;


	case BLE_GAP_EVT_ADV_REPORT:
	{
		++m_totalAdvertisementsReceived;
		data_t uuidData;// if found, this points into the advertised data we received, so we don't 'own' the data
		char strUUID[UUID_CHARACTER_LENGTH_NONULL + 1/*null*/];
		if (find_adv_uuid(&p_ble_evt->evt.gap_evt.params.adv_report, uuidData)
			&& (UUID_to_string_convert(uuidData.data_len, uuidData.p_data, strUUID, sizeof(strUUID) / sizeof(strUUID[0])))
			)
		{
			ble_gap_addr_t peerAddr(p_ble_evt->evt.gap_evt.params.adv_report.peer_addr);
			int8_t rssi = p_ble_evt->evt.gap_evt.params.adv_report.rssi;
			std::string stdStrUUID(strUUID);
			message_queue_.Post([peerAddr, stdStrUUID, rssi]() {
				try {
					on_adv_report(peerAddr, stdStrUUID, rssi);
				}
				catch (const std::exception& ex)
				{
					printf("caught exception from on_adv_report: %s\n", ex.what());
				}
				catch (...)
				{
					printf("caught non-std::exception from on_adv_report\n");
				}
			});
		}
		else
		{
			if (m_iLoggingLevel >= 6)
				printf("advertisement received that does not have 128-bit UUID\n");
		}
		message_queue_.Post([]() {
			try {
				if (!m_DeviceBeingDiscovered)
				{
					scan_restart();//re-start scanning after receiving advertisement
				}
			}
			catch (const std::exception& ex)
			{
				printf("caught exception from scan re-start: %s\n", ex.what());
			}
			catch (...)
			{
				printf("caught non-std::exception from scan re-start\n");
			}
		});
		}
		break;

	case BLE_GAP_EVT_TIMEOUT:
		on_ble_gap_evt_timeout(&(p_ble_evt->evt.gap_evt));
		break;
	case BLE_GATTC_EVT_TIMEOUT:
		on_gattc_timeout();// The only extra information for the timeout is p_ble_evt->evt.gattc_evt.params.timeout.src, which has only one valid value BLE_GATT_TIMEOUT_SRC_PROTOCOL (the only value for BLE_GATT_TIMEOUT_SOURCES)
		break;

	case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
		on_service_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_CHAR_DISC_RSP:
		on_characteristic_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_DESC_DISC_RSP:
		on_descriptor_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_WRITE_RSP:
		on_write_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_HVX:
		on_hvx(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
		on_conn_params_update_request(&(p_ble_evt->evt.gap_evt));
		break;

#if NRF_SD_BLE_API >= 3
#if 0
	case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
		on_exchange_mtu_request(&(p_ble_evt->evt.gatts_evt));
		break;

	case BLE_GATTC_EVT_EXCHANGE_MTU_RSP:
		on_exchange_mtu_response(&(p_ble_evt->evt.gattc_evt));
		break;
#endif
#endif
	case BLE_GAP_EVT_RSSI_CHANGED:
		update_connected_device_rssi_changed(p_ble_evt->evt.gap_evt.conn_handle, p_ble_evt->evt.gap_evt.params.rssi_changed.rssi);
		if(m_iLoggingLevel >= 7)
		{
			printf("connection %d RSSI = %d (channel %d)\n", p_ble_evt->evt.gap_evt.conn_handle, p_ble_evt->evt.gap_evt.params.rssi_changed.rssi, p_ble_evt->evt.gap_evt.params.rssi_changed.ch_index);
		}
		break;

	case BLE_GATTC_EVT_READ_RSP:
		on_gattc_evt_read_response(&p_ble_evt->evt.gattc_evt);
		break;
	case BLE_GATTS_EVT_WRITE:
		printf("%s: BLE_GATTS_EVT_WRITE not handled\n", GetCurrentDateTimeString().c_str());
		break;
	case BLE_GAP_EVT_ADV_SET_TERMINATED:
		// This means the buffer we passed to sd_ble_gap_adv_set_configure is no longer being used.
		// Since we currently set advertising to continue indefinitely, we should not get this.
		printf("BLE_GAP_EVT_ADV_SET_TERMINATED event is not handled, adv_handle = %d\n", p_ble_evt->evt.gap_evt.params.adv_set_terminated.adv_handle);
		break;
	case BLE_GAP_EVT_SCAN_REQ_REPORT:
		printf("BLE_GAP_EVT_SCAN_REQ_REPORT event is not handled\n");
		break;
	case BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE:
		if(m_iLoggingLevel >= 4)
			printf("%s: BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE received. Expected after challenge sent.\n", GetCurrentDateTimeString().c_str());
		break;
	case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
		// This event happens after every connection
		if(m_iLoggingLevel >= 6)
			printf("BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST received. No action taken.\n");
		break;
	default:
		printf("%s: Received an un-handled event with ID: %d\n", GetCurrentDateTimeString().c_str(), p_ble_evt->header.evt_id);
		break;
	}
	fflush(stdout);
}

bool ReadFeatureCharacteristicOfOpenConnections()
{
	bool bSuccess = false;
	{
		if (m_DeviceBeingDiscovered)
			return false;

		// Try to contact every open connection
		struct ConnectionAndCharacteristic {
			ConnectedDevice* m_conn;
			uint16_t m_charHandle;

			ConnectionAndCharacteristic(ConnectedDevice* con, uint16_t ch) : m_conn(con), m_charHandle(ch) {}
		};
		std::vector< ConnectionAndCharacteristic> openConnections;

		for (auto dev : ConnectedDevices)
		{
			if (dev.second->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
			{
				bool bFoundFeatureCharacteristic = false;// found the Imprivata ID 'feature' characteristic, whose UUID is known to us
				for (auto srv : dev.second->m_Services)
				{
					if (srv.m_uuid_string == dev.second->m_UUID)
					{
						for (auto ch : srv.m_Characteristics)
						{
							// for M770 was hacked to check for if(ch.m_ValueHandle != 0x42)
							if (_stricmp(ch.m_ValueHandleUUID.c_str(), IMPRIVATA_ID_FEATURE_CHARACTERISTIC_UUID_NO_DASHES) != 0)
								continue;
							if (m_iLoggingLevel >= 4)
								printf("authentication service %s 'feature' characteristic value 0x%04X handle\n",
									srv.m_uuid_string.c_str(),
									ch.m_ValueHandle);
							fflush(stdout);
							openConnections.push_back(ConnectionAndCharacteristic(dev.second, ch.m_ValueHandle));
							bFoundFeatureCharacteristic = true;
							break;
						}
					}
				}
				if (bFoundFeatureCharacteristic == false)
				{
					printf("phone %s connection_handle = %d could not find 'feature' characteristic\n",
						dev.second->m_UUID.c_str(),
						dev.second->m_ConnectionHandle);
					fflush(stdout);
				}
			}
		}
		for (auto con : openConnections)
		{
			if (hEvent_current_GATTC)
			{
				printf("%s: WARNING: skipping sd_ble_gattc_read when there is a GATTC operation already in progress\n",
					GetCurrentDateTimeString().c_str());
				break;
			}

			current_GATTC_timed_out = false;
			hEvent_current_GATTC = hEvent_BLE_GATTC_EVT_READ_RSP;

			// To keep the connection alive we can do anything at all that generates traffic to the phone
			// This will trigger event BLE_GATTC_EVT_READ_RSP (on_gattc_evt_read_response)
			::ResetEvent(hEvent_BLE_GATTC_EVT_READ_RSP);// we don't wait for this during service discovery, make sure it starts non-signalled.


			uint32_t err_code;
			{
				std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);

				err_code = sd_ble_gattc_read(m_adapter, con.m_conn->m_ConnectionHandle, con.m_charHandle, 0/*offset*/);
			}
			if (err_code != NRF_SUCCESS)
			{
				hEvent_current_GATTC = 0;// since we won't wait on event
				printf("Keepalive sd_ble_gattc_read error for connection %d characteristic value handle 0x%04X error %s\n",
					con.m_conn->m_ConnectionHandle,
					con.m_charHandle,
					bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
			}
			else
			{
				// We should get event BLE_GATTC_EVT_READ_RSP.
				// The maximum time according to the BLE spec is 30 seconds, after which pc-ble-driver gives us a timeout event.
				const DWORD millisecondsToWaitForReadEvent = 65000;
				DWORD dWait = ::WaitForSingleObject(hEvent_BLE_GATTC_EVT_READ_RSP, millisecondsToWaitForReadEvent);
				hEvent_current_GATTC = 0;// done waiting on event
				if (dWait == WAIT_OBJECT_0)
				{
					if (current_GATTC_timed_out)
					{
						printf("%s: sd_ble_gattc_read to keep connection alive got BLE_GATTC_EVT_TIMEOUT or BLE_GAP_EVT_DISCONNECTED (discard this connection and re-connect)\n", GetCurrentDateTimeString().c_str());
						if(con.m_conn->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)// check, just in case we got a disconnect event
							force_disconnect(con.m_conn->m_ConnectionHandle, true);
					}
					else
					{
						bSuccess = true;
						if (m_iLoggingLevel >= 4)
						{
							printf("%s: Thread %d: got expected BLE_GATTC_EVT_READ_RSP\n",
								GetCurrentDateTimeString().c_str(),
								::GetCurrentThreadId()
							);
						}
					}
				}
				else if (dWait == WAIT_TIMEOUT)
				{
					printf("%s: Thread %d: ERROR: never got BLE_GATTC_EVT_READ_RSP (or a timeout event) after waiting %d milliseconds.\n",
						GetCurrentDateTimeString().c_str(),
						::GetCurrentThreadId(),
						millisecondsToWaitForReadEvent
					);
					if (con.m_conn->m_ConnectionHandle != BLE_CONN_HANDLE_INVALID)
					{
						printf("Before exiting, disconnect and see if we get a disconnect event\n");
						force_disconnect(con.m_conn->m_ConnectionHandle, true/*do wait for disconnect event, we want to know if it exists*/);
					}
					printf("Exit test program immediately\n");
					::ExitProcess(0/*exit code*/);
				}
				else
				{
					printf("ERROR: WaitForSingleObject returned unexpected value %d while waiting for BLE_GATTC_EVT_READ_RSP\n", dWait);
				}
			}

		}
	}
	return bSuccess;
}
// Thread to keep connections alive by polling them
void __stdcall KeepAliveThread(void* argList)
{
	if (m_DoScanning == false)
		return;// if we're not scanning, there are no connections to keep alive

	HANDLE hEvent = (HANDLE)argList;
	bool bEnable = true;
	unsigned nChallengeResponsesSinceDisconnect = 0;

	while(::WaitForSingleObject(hEvent, 3000/*3 seconds*/) == WAIT_TIMEOUT)
	{
#if 1
		// If we have just connected to a device and are finding its services, don't do this keepalive logic. Wait until the connection is established.
		if (m_DeviceBeingDiscovered)
			continue;

		if (ReadFeatureCharacteristicOfOpenConnections())// if this fails, don't bother with challenge/response
		{
			do_challenge_response_to_connected_phones();
			++nChallengeResponsesSinceDisconnect;
			if (nChallengeResponsesSinceDisconnect > 10)
			{
				nChallengeResponsesSinceDisconnect = 0;
				if (m_iLoggingLevel >= 5)
					printf("Force disconnect on all open connections, so we can go through connection process again.\n");
				force_disconnect_on_all_connections();
			}
		}
		else
			force_disconnect_on_all_connections();

		// I've seen scan not restarted, make sure we eventually do restart it
		unsigned long long timestampNow = ::GetTickCount64();
		if((timestampNow - m_timestampLastAdvertisement > 30000) && (timestampNow - m_timestampLastScanStart > 30000))
		{
			if(m_iLoggingLevel >= 1)
				printf("%s: Re-start scanning due to no advertisements in a long time\n",
					GetCurrentDateTimeString().c_str()
					);
			{
				std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
				::sd_ble_gap_scan_stop(m_adapter);// stop scanning before trying to re-start in keepalive thread because we've received no advertisements lately
			}
			scan_restart();//re-start in keepalive thread because we've received no advertisements lately
		}
#else
		// can we keep other connections alive while one is being discovered? I sure hope so, otherwise they may time out
		if(m_DeviceBeingDiscovered)
			continue;

		// Try to contact every open connection
		struct ConnectionAndCharacteristic {
			std::string m_serviceUUID;
			uint16_t m_conn;
			uint16_t m_charHandle;

			ConnectionAndCharacteristic(std::string serviceUuid, uint16_t con, uint16_t ch) : m_serviceUUID(serviceUuid), m_conn(con), m_charHandle(ch) {}
		};
		std::vector< ConnectionAndCharacteristic> openConnections;

		for(auto dev : ConnectedDevices)
		{
			auto conn = dev.second->m_ConnectionHandle;
			if(conn != BLE_CONN_HANDLE_INVALID)
			{
				bool bFoundCharacteristic = false;// found the Imprivata ID 'feature' characteristic, whose UUID is known to us
				for(auto srv : dev.second->m_Services)
				{
					if(srv.m_uuid_string == dev.second->m_UUID)
					{
						for(auto ch : srv.m_Characteristics)
						{
							if(_stricmp(ch.m_ValueHandleUUID.c_str(), IMPRIVATA_ID_RESPONSE_CHARACTERISTIC_UUID_NO_DASHES) != 0)
								continue;
							if(m_iLoggingLevel >= 4)
								printf("authentication service %s 'response' characteristic value 0x%04X handle\n", 
									srv.m_uuid_string.c_str(),
									ch.m_ValueHandle);
							fflush(stdout);
							openConnections.push_back(ConnectionAndCharacteristic(dev.second->m_UUID, conn, ch.m_ValueHandle));
							bFoundCharacteristic = true;
							break;
						}
					}
				}
				if(bFoundCharacteristic == false)
				{
					printf("phone %s connection_handle = %d could not find the 'response' characteristic for challenge/response\n",
						dev.second->m_UUID.c_str(),
						dev.second->m_ConnectionHandle);
					fflush(stdout);
				}
			}
		}
		for(auto con : openConnections)
		{
			// To keep the connection alive we can do anything at all that generates traffic to the phone
			enable_notifications_for_response_characteristic(con.m_serviceUUID, con.m_conn, con.m_charHandle, bEnable);
			bEnable = !bEnable;
		}
#endif
	}
	// event got signaled, time to exit
	printf("KeepAliveThread exiting\n"); fflush(stdout);
}
// Thread to keep cycling advertisements by changing the UUID and/or the major/minor fields
void __stdcall AdvertisingThread(void* argList)
{
	HANDLE hEvent = (HANDLE)argList;
	bool bFirstTime = true;
	while (::WaitForSingleObject(hEvent, 3000/*3 seconds*/) == WAIT_TIMEOUT)
	{
		if (Calls_to_pc_ble_driver::m_bSilentNoBeacon)
		{
			printf("Option specified to not beacon, exiting the thread\n");
			return;
		}
		if(true)
		{
			if(bFirstTime == false)
			{
				advertising_stop();
			}
			advertising_data_setup();// run through init again
			advertising_start();
			bFirstTime = false;
		}
	}
	// event got signaled, time to exit
	printf("AdvertisingThread exiting\n"); fflush(stdout);
	fflush(stdout);
}



static void do_hard_reset_of_Nordic_device()
{
	uint32_t err_code = ::sd_rpc_conn_reset(m_adapter, SYS_RESET/*hard reset that causes USB device removal then arrival*/);

	if(err_code != NRF_SUCCESS)
	{
		printf("sd_rpc_conn_reset failed error %s", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
	}
}

/** Main */

/**@brief Function for application main entry.
 *
 * @param[in] argc Number of arguments (program expects 0 or 1 arguments).
 * @param[in] argv The serial port of the target nRF5 device (Optional).
 */

int Calls_to_pc_ble_driver::scanner_main()
{
	m_iLoggingLevel = m_iLoggingLevel_;
	uint32_t err_code;
	uint32_t baud_rate = DEFAULT_BAUD_RATE;
	const char* serial_port = m_serial_port_.c_str();

	for(auto uuid : m_uuidsToFind_)
	{
		if (uuid.length() != UUID_CHARACTER_LENGTH_NONULL)
		{
			printf("UUID %s does not have expected %d characters\n", uuid.c_str(), UUID_CHARACTER_LENGTH_NONULL);
			return NRF_ERROR_INTERNAL;
		}
		ConnectedDevices.insert(std::make_pair(uuid, new ConnectedDevice(uuid)));
	}


	if(m_iLoggingLevel >=2)
		printf("Serial port used: %s\n", serial_port);
	if (m_iLoggingLevel >= 2)
		printf("Baud rate used: %d\n", baud_rate);
	printf("'c' to do a challenge/response to connected phones (the response is not validated)\n");
	printf("'d' to disconnect any connected devices\n");
	printf("'q' to quit program\n");
	printf("'r' to do a hard reset of the dongle - after this the program exits\n");
	printf("'s' to show status of all monitored phones\n");
	printf("'+' to increase logging level\n");
	printf("'-' to decrease logging level\n");
	fflush(stdout);

	m_adapter = adapter_init(const_cast<char*>(serial_port), baud_rate);
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		sd_rpc_log_handler_severity_filter_set(m_adapter, SD_RPC_LOG_INFO);
	}
	m_timestampOpenNordicDevice = ::GetTickCount64();
	{
		std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
		err_code = sd_rpc_open(m_adapter, status_handler, ble_evt_dispatch, log_handler);
	}

	if (err_code != NRF_SUCCESS)
	{
		printf("Failed to open nRF BLE Driver. Error %s\n", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
		fflush(stdout);
		return err_code;
	}

#if NRF_SD_BLE_API >= 5
	ble_cfg_set(m_config_id);
#endif

	err_code = ble_stack_init();

	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}

#if NRF_SD_BLE_API < 5
	err_code = ble_options_set();

	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}
#endif

	err_code = scan_start();// initial scan start after opening device

	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}


	hEvent_BLE_GATTC_EVT_WRITE_RSP = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		false/*bManualReset=false, so event is reset by WaitForSingleObject*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hEvent_BLE_GATTC_EVT_WRITE_RSP == 0)
	{
		printf("failed to create event for receiving BLE_GATTC_EVT_WRITE_RSP. error = %d\n", ::GetLastError());
		return -1;
	}

	hEvent_BLE_GATTC_EVT_READ_RSP = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		false/*bManualReset=false, so event is reset by WaitForSingleObject*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hEvent_BLE_GATTC_EVT_READ_RSP == 0)
	{
		printf("failed to create event for receiving BLE_GATTC_EVT_READ_RSP. error = %d\n", ::GetLastError());
		return -1;
	}

	hEvent_BLE_GAP_EVT_DISCONNECTED = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		true/*bManualReset=true, so event is reset by calls to ResetEvent, not WaitForSingleObject*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hEvent_BLE_GAP_EVT_DISCONNECTED == 0)
	{
		printf("failed to create event to for receiving BLE_GAP_EVT_DISCONNECTED. error = %d\n", ::GetLastError());
		return -1;
	}

	hEvent_BLE_GAP_EVT_CONNECT_Completed = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		true/*bManualReset=true, so event is reset by calls to ResetEvent, not WaitForSingleObject*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hEvent_BLE_GAP_EVT_CONNECT_Completed == 0)
	{
		printf("failed to create event to for receiving BLE_GAP_EVT_DISCONNECTED. error = %d\n", ::GetLastError());
		return -1;
	}

	HANDLE hStopKeepAliveThreadEvent = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		true/*bManualReset*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hStopKeepAliveThreadEvent == 0)
	{
		printf("failed to create event to stop the keep-alive thread. error = %d\n", ::GetLastError());
		return -1;
	}
	HANDLE threadHandleKeepAlive = (HANDLE)_beginthread( // NATIVE CODE
		KeepAliveThread,
		0/*stack_size, use default*/,
		hStopKeepAliveThreadEvent/*argument to thread function*/
	);
	HANDLE hStopAdvertisingThreadEvent = CreateEventA(
		(LPSECURITY_ATTRIBUTES)nullptr,
		true/*bManualReset*/,
		false/*bInitialState*/,
		nullptr/*Name*/
	);
	if (hStopAdvertisingThreadEvent == 0)
	{
		printf("failed to create event to stop the advertising thread. error = %d\n", ::GetLastError());
		return -1;
	}

	HANDLE threadHandleAdvertising = (HANDLE)_beginthread( // NATIVE CODE
		AdvertisingThread,
		0/*stack_size, use default*/,
		hStopAdvertisingThreadEvent/*argument to thread function*/
	);

	// Endlessly loop.
	for (;;)
	{
		char c = (char)getchar();
		if (c == 'q' || c == 'Q')
		{
			m_bShuttingDown = true;
			
			::SetEvent(hStopKeepAliveThreadEvent);// tell keepalive thread to exit
			printf("Wait for keep-alive thread to stop\n");
			::WaitForSingleObject(threadHandleKeepAlive, INFINITE);// wait for keepalive thread to exit
			::SetEvent(hStopAdvertisingThreadEvent);// tell Advertising thread to exit
			printf("Wait for advertising thread to stop\n");
			::WaitForSingleObject(threadHandleAdvertising, INFINITE);// wait for Advertising thread to exit

			printf("Stop advertising\n");
			advertising_stop();// If we don't stop advertising, it will keep advertising after this test program is done
			printf("Call sd_rpc_close\n");
			{
				std::unique_lock<std::mutex> lck(m_MutexEnforceSingleCallToNordic);
				err_code = sd_rpc_close(m_adapter);
			}

			if (err_code != NRF_SUCCESS)
			{
				printf("Failed to close nRF BLE Driver. Error %s\n", bgsdk::Nordic::NordicErrorCodeToMessage(err_code).c_str());
				fflush(stdout);
			}
			else
			{
				printf("Closed\n");
				fflush(stdout);
			}
			break;
		}
		else if(c == 'c')
		{
			do_challenge_response_to_connected_phones();
		}
		else if(c == 'r')
		{
			do_hard_reset_of_Nordic_device();
			break;
		}
		else if(c == 's')
		{
			double advertsPerSecond = (m_totalAdvertisementsReceived / ((::GetTickCount64() - m_timestampOpenNordicDevice)/ 1000.0));
			printf("%s: total advertisements received = %d (%f/second); total 128-bit UUID advertisements = %d\n",
				GetCurrentDateTimeString().c_str(),
				m_totalAdvertisementsReceived,
				advertsPerSecond,
				m_totalAdvertisementsReceivedWith128BitUUIDs);
			log_connected_device_status();
			if (m_iLoggingLevel >= 4)
				log_all_128bitUUID_devices_seen();
		}
		else if (c == '+')
		{
			++m_iLoggingLevel;
			printf("logging level %d\n", m_iLoggingLevel);
		}
		else if (c == '-')
		{
			if (m_iLoggingLevel >= 1)
			{
				--m_iLoggingLevel;
			}
			printf("logging level %d\n", m_iLoggingLevel);
		}
		else if(c == 'd')
		{
			force_disconnect_on_all_connections();
		}
	}
	message_queue_.StopAndWaitThreadEnd();
	printf("Exiting\n");
	return err_code;
}



Calls_to_pc_ble_driver::Calls_to_pc_ble_driver(const std::string& serial_port
	, const std::set<std::string>& uuidsToFind
	, const int iLoggingLevel/*0 little, 1 more, 2 a lot*/
	, const bool bSilentNoBeacon/*if true, do not beacon*/
	, uint16_t scan_interval/*pass 0 to get default*/
	, uint16_t scan_window/*pass 0 to get default*/) :
	m_serial_port_(serial_port),
	m_uuidsToFind_(uuidsToFind),
	m_iLoggingLevel_(iLoggingLevel)
{
	m_DoScanning = !m_uuidsToFind_.empty();
	m_scan_interval = scan_interval;
	m_scan_window = scan_window;
	m_bSilentNoBeacon = bSilentNoBeacon;
}


Calls_to_pc_ble_driver::~Calls_to_pc_ble_driver()
{
}


ble_gap_scan_params_t Calls_to_pc_ble_driver::GetScanParams()
{
	ble_gap_scan_params_t scan_param =
	{
	#if NRF_SD_BLE_API >= 6
		0,                       // Set if accept extended advertising packetets.
		0,                       // Set if report inomplete reports.
	#endif
		0,                       // Set if active scanning.
	#if NRF_SD_BLE_API < 6
		0,                       // Set if selective scanning.
	#endif
	#if NRF_SD_BLE_API >= 6
		BLE_GAP_SCAN_FP_ACCEPT_ALL,
		BLE_GAP_PHY_1MBPS,
	#endif
	#if NRF_SD_BLE_API == 2
		NULL,                    // Set white-list.
	#endif
	#if NRF_SD_BLE_API == 3 || NRF_SD_BLE_API == 5
		0,                       // Set adv_dir_report.
	#endif
		(uint16_t)SCAN_INTERVAL,
		(uint16_t)SCAN_WINDOW,
		(uint16_t)SCAN_TIMEOUT
	#if NRF_SD_BLE_API >= 6
		, { 0 }                  // Set chennel mask.
	#endif
	};
	if (m_scan_interval != 0)
		scan_param.interval = m_scan_interval;
	if (m_scan_window != 0)
		scan_param.window = m_scan_window;
	return scan_param;
}

/*static*/ bool					Calls_to_pc_ble_driver::m_bSilentNoBeacon = false;/*if true, do not beacon*/
/*static*/ uint16_t				Calls_to_pc_ble_driver::m_scan_interval = 0;/*pass 0 to get default*/
/*static*/ uint16_t				Calls_to_pc_ble_driver::m_scan_window = 0;/*pass 0 to get default*/